import pytest
import responses

import globus_sdk
from globus_sdk.testing import get_last_request, load_response

IDENTITIES_MULTIPLE_RESPONSE = {
    "identities": [
        {
            "email": None,
            "id": "46bd0f56-e24f-11e5-a510-131bef46955c",
            "identity_provider": "927d7238-f917-4eb2-9ace-c523fa9ba34e",
            "name": None,
            "organization": None,
            "status": "unused",
            "username": "globus@globus.org",
        },
        {
            "email": "sirosen@globus.org",
            "id": "ae341a98-d274-11e5-b888-dbae3a8ba545",
            "identity_provider": "927d7238-f917-4eb2-9ace-c523fa9ba34e",
            "name": "Stephen Rosen",
            "organization": "Globus Team",
            "status": "used",
            "username": "sirosen@globus.org",
        },
    ]
}


def test_identity_map(service_client):
    meta = load_response(service_client.get_identities, case="sirosen").metadata
    idmap = globus_sdk.IdentityMap(service_client, [meta["username"]])
    assert idmap[meta["username"]]["organization"] == meta["org"]

    # lookup by ID also works
    assert idmap[meta["id"]]["organization"] == meta["org"]

    # the last (only) API call was the one by username
    last_req = get_last_request()
    assert "ids" not in last_req.params
    assert last_req.params == {"usernames": meta["username"], "provision": "false"}


def test_identity_map_initialization_no_values(service_client):
    idmap = globus_sdk.IdentityMap(service_client)
    assert idmap.unresolved_ids == set()
    assert idmap.unresolved_usernames == set()


def test_identity_map_initialization_mixed_and_duplicate_values(service_client):
    # splits things up and deduplicates values into sets
    idmap = globus_sdk.IdentityMap(
        service_client,
        [
            "sirosen@globus.org",
            "ae341a98-d274-11e5-b888-dbae3a8ba545",
            "globus@globus.org",
            "sirosen@globus.org",
            "globus@globus.org",
            "ae341a98-d274-11e5-b888-dbae3a8ba545",
            "sirosen@globus.org",
            "ae341a98-d274-11e5-b888-dbae3a8ba545",
        ],
    )
    assert idmap.unresolved_ids == {"ae341a98-d274-11e5-b888-dbae3a8ba545"}
    assert idmap.unresolved_usernames == {"sirosen@globus.org", "globus@globus.org"}


def test_identity_map_initialization_batch_size(service_client):
    idmap = globus_sdk.IdentityMap(service_client, id_batch_size=10)
    assert idmap.unresolved_ids == set()
    assert idmap.unresolved_usernames == set()
    assert idmap.id_batch_size == 10


def test_identity_map_add(service_client):
    idmap = globus_sdk.IdentityMap(service_client)
    assert idmap.add("sirosen@globus.org") is True
    assert idmap.add("sirosen@globus.org") is False
    assert idmap.add("46bd0f56-e24f-11e5-a510-131bef46955c") is True
    assert idmap.add("46bd0f56-e24f-11e5-a510-131bef46955c") is False


def test_identity_map_add_after_lookup(service_client):
    meta = load_response(service_client.get_identities, case="sirosen").metadata
    idmap = globus_sdk.IdentityMap(service_client)
    x = idmap[meta["username"]]["id"]
    # this is the key: adding it will indicate that we've already seen this ID, perhaps
    # "unintuitively", and that's part of the value of `add()` returning a boolean value
    assert idmap.add(x) is False
    assert idmap[x] == idmap[meta["username"]]


def test_identity_map_multiple(service_client):
    meta = load_response(service_client.get_identities, case="multiple").metadata
    idmap = globus_sdk.IdentityMap(
        service_client, ["sirosen@globus.org", "globus@globus.org"]
    )
    assert idmap["sirosen@globus.org"]["organization"] == "Globus Team"
    assert idmap["globus@globus.org"]["organization"] is None

    last_req = get_last_request()
    # order doesn't matter, but it should be just these two
    # if IdentityMap doesn't deduplicate correctly, it could send
    # `sirosen@globus.org,globus@globus.org,sirosen@globus.org` on the first lookup
    assert last_req.params["usernames"].split(",") in [
        meta["usernames"],
        meta["usernames"][::-1],
    ]
    assert last_req.params["provision"] == "false"


def test_identity_map_keyerror(service_client):
    load_response(service_client.get_identities, case="sirosen")
    idmap = globus_sdk.IdentityMap(service_client)
    # a name which doesn't come back, indicating that it was not found, will KeyError
    with pytest.raises(KeyError):
        idmap["sirosen2@globus.org"]

    last_req = get_last_request()
    assert last_req.params == {"usernames": "sirosen2@globus.org", "provision": "false"}


def test_identity_map_get_with_default(service_client):
    load_response(service_client.get_identities, case="sirosen")
    magic = object()  # sentinel value
    idmap = globus_sdk.IdentityMap(service_client)
    # a name which doesn't come back, if looked up with `get()` should return the
    # default
    assert idmap.get("sirosen2@globus.org", magic) is magic


def test_identity_map_del(service_client):
    meta = load_response(service_client.get_identities).metadata
    idmap = globus_sdk.IdentityMap(service_client)
    identity_id = idmap[meta["username"]]["id"]
    del idmap[identity_id]
    assert idmap.get(meta["username"])["id"] == identity_id
    # we've only made one request so far
    assert len(responses.calls) == 1
    # but a lookup by ID after a del is going to trigger another request because we've
    # invalidated the cached ID data and are asking the IDMap to look it up again
    assert idmap.get(identity_id)["username"] == meta["username"]
    assert len(responses.calls) == 2


@pytest.mark.parametrize(
    "lookup1,lookup2",
    [
        ("username", "username"),
        ("username", "id"),
        ("id", "username"),
    ],
)
@pytest.mark.parametrize("initial_add", [True, False])
def test_identity_map_shared_cache_match(service_client, initial_add, lookup1, lookup2):
    meta = load_response(service_client.get_identities, case="sirosen").metadata
    lookup1, lookup2 = meta[lookup1], meta[lookup2]
    cache = {}
    idmap1 = globus_sdk.IdentityMap(service_client, cache=cache)
    idmap2 = globus_sdk.IdentityMap(service_client, cache=cache)
    if initial_add:
        idmap1.add(lookup1)
        idmap2.add(lookup2)
    # no requests yet...
    assert len(responses.calls) == 0
    # do the first lookup, it should make one request
    assert idmap1[lookup1]["organization"] == meta["org"]
    assert len(responses.calls) == 1
    # lookup more values and make sure that "everything matches"
    assert idmap2[lookup2]["organization"] == meta["org"]
    assert idmap1[lookup1]["id"] == idmap2[lookup2]["id"]
    assert idmap1[lookup2]["id"] == idmap2[lookup1]["id"]
    # we've only made one request, because the shared cache captured this info on the
    # very first call
    assert len(responses.calls) == 1


@pytest.mark.parametrize(
    "lookup_style",
    ["usernames", "ids"],
)
def test_identity_map_shared_cache_mismatch(service_client, lookup_style):
    meta = load_response(service_client.get_identities, case="multiple").metadata
    lookup1, lookup2 = meta[lookup_style]
    cache = {}
    idmap1 = globus_sdk.IdentityMap(service_client, [lookup1, lookup2], cache=cache)
    idmap2 = globus_sdk.IdentityMap(service_client, cache=cache)
    # no requests yet...
    assert len(responses.calls) == 0
    # do the first lookup, it should make one request
    assert lookup1 in (idmap1[lookup1]["id"], idmap1[lookup1]["username"])
    assert len(responses.calls) == 1
    # lookup more values and make sure that "everything matches"
    assert lookup2 in (idmap2[lookup2]["id"], idmap2[lookup2]["username"])
    assert idmap1[lookup1]["id"] == idmap2[lookup1]["id"]
    assert idmap1[lookup2]["id"] == idmap2[lookup2]["id"]
    # we've only made one request, because the shared cache captured this info on the
    # very first call
    assert len(responses.calls) == 1


def test_identity_map_prepopulated_cache(service_client):
    meta = load_response(service_client.get_identities).metadata

    # populate the cache, even with nulls it should stop any lookups from happening
    cache = {meta["id"]: None, meta["username"]: None}
    idmap = globus_sdk.IdentityMap(service_client, cache=cache)
    # no requests yet...
    assert len(responses.calls) == 0
    # do the lookups
    assert idmap[meta["id"]] is None
    assert idmap[meta["username"]] is None
    # still no calls made
    assert len(responses.calls) == 0


def test_identity_map_batch_limit(service_client):
    meta1 = load_response(service_client.get_identities).metadata
    meta2 = load_response(service_client.get_identities, case="sirosen").metadata

    # setup the ID map with a size limit of 1
    idmap = globus_sdk.IdentityMap(service_client, id_batch_size=1)
    idmap.add(meta2["id"])
    idmap.add(meta1["id"])

    # no requests yet...
    assert len(responses.calls) == 0

    # do the first lookup, using the second ID to be added
    # only one call should be made
    assert idmap[meta1["id"]]["username"] == meta1["username"]
    assert len(responses.calls) == 1
    # 1 ID left unresolved
    assert len(idmap.unresolved_ids) == 1
    # the last (only) API call was by ID with one ID
    last_req = get_last_request()
    assert "usernames" not in last_req.params
    assert last_req.params == {"ids": meta1["id"]}

    # second lookup works as well
    assert idmap[meta2["id"]]["username"] == meta2["username"]
    assert len(responses.calls) == 2
    # no IDs left unresolved
    assert len(idmap.unresolved_ids) == 0
    # the last API call was by ID with one ID
    last_req = get_last_request()
    assert "usernames" not in last_req.params
    assert last_req.params == {"ids": meta2["id"]}



import pytest

from globus_sdk._internal.utils import get_nice_hostname, sha256_string, slash_join


def test_sha256string():
    test_string = "foo"
    expected_sha = "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae"

    assert sha256_string(test_string) == expected_sha


@pytest.mark.parametrize(
    "platform_value, result",
    (
        # platform.node() can return '' when it doesn't know the hostname
        # turn this into None
        pytest.param("", None, id="empty-is-none"),
        # macOS adds '.local' to the user's chosen machine name
        pytest.param(
            "VeryCoolMacbook.local", "VeryCoolMacbook", id="remove-local-suffix"
        ),
        # the "boring" case is when we do no extra work
        pytest.param("linux-workstation", "linux-workstation", id="boring"),
    ),
)
def test_get_nice_hostname(platform_value, result, monkeypatch):
    monkeypatch.setattr("platform.node", lambda: platform_value)
    assert get_nice_hostname() == result


@pytest.mark.parametrize(
    "a, b",
    [(a, b) for a in ["a", "a/"] for b in ["b", "/b"]]
    + [("a/b", c) for c in ["", None]],  # type: ignore
)
def test_slash_join(a, b):
    """
    slash_joins a's with and without trailing "/"
    to b's with and without leading "/"
    Confirms all have the same correct slash_join output
    """
    assert slash_join(a, b) == "a/b"



"""
The definition of the MISSING sentinel and its type.

These are exposed publicly as `globus_sdk.MISSING` and `globus_sdk.MissingType`.
"""

from __future__ import annotations

import typing as t

T = t.TypeVar("T")

# type checkers don't know that MISSING is a sentinel, so we will describe it
# differently at typing time, allowing for type narrowing on `is MISSING` and
# similar checks
if t.TYPE_CHECKING:
    # pretend that `MISSING: MissingType` is an enum at type-checking time
    #
    # enums are treated as `Literal[...]` values and are narrowed under simple
    # checks, as unions and literal types are
    # therefore, under this definition, `MissingType ~= Literal[MissingType.MISSING]`
    #
    # Therefore, consider this example:
    #
    #       x: int | float | MissingType
    #       if x is not MISSING:
    #           reveal_type(x)
    #
    # This is effectively the same as if we wrote:
    #
    #       x: int | float | Literal["a"]
    #       if x != "a":
    #           reveal_type(x)
    #
    # Both should show `x: int | float`
    import enum

    class MissingType(enum.Enum):
        MISSING = enum.auto()

    MISSING = MissingType.MISSING
else:

    class MissingType:
        def __init__(self) -> None:
            # disable instantiation, but gated to be able to run once
            # when this module is imported
            if "MISSING" in globals():
                raise TypeError("MissingType should not be instantiated")

        def __bool__(self) -> bool:
            return False

        def __copy__(self) -> MissingType:
            return self

        def __deepcopy__(self, memo: dict[int, t.Any]) -> MissingType:
            return self

        # unpickling a MissingType should always return the "MISSING" sentinel
        def __reduce__(self) -> str:
            return "MISSING"

        def __repr__(self) -> str:
            return "<globus_sdk.MISSING>"

    # a sentinel value for "missing" values which are distinguished from `None` (null)
    # this is the default used to indicate that a parameter was not passed, so that
    # method calls passing `None` can be distinguished from those which did not pass any
    # value
    # users should typically not use this value directly, but it is part of the
    # public SDK interfaces along with its type for annotation purposes
    MISSING = MissingType()


@t.overload
def filter_missing(data: dict[str, t.Any]) -> dict[str, t.Any]: ...


@t.overload
def filter_missing(data: None) -> None: ...


def filter_missing(data: dict[str, t.Any] | None) -> dict[str, t.Any] | None:
    if data is None:
        return None
    return {k: v for k, v in data.items() if v is not MISSING}


def none2missing(obj: T | None) -> T | MissingType:
    """
    A converter for interfaces which take "nullable" to mean "omittable", to
    adapt them to usage sites which require use of MISSING for omittable
    elements.

    :param obj: The nullable object to convert to an omittable object.
    """
    if obj is None:
        return MISSING
    return obj



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID

DEFAULT_RESPONSE_DOC = {
    "status": "online",
    "details": {
        "total_workers": 1,
        "idle_workers": 0,
        "pending_tasks": 0,
        "outstanding_tasks": 0,
        "managers": 1,
    },
}

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/endpoints/{ENDPOINT_ID}/status",
        method="GET",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



"""
Custom retry check collection for the TransferClient that overrides
the default check_transient_error
"""

from __future__ import annotations

from globus_sdk.transport import RetryCheck, RetryCheckResult, RetryContext
from globus_sdk.transport.default_retry_checks import (
    DEFAULT_RETRY_CHECKS,
    check_transient_error,
)


def check_transfer_transient_error(ctx: RetryContext) -> RetryCheckResult:
    """
    check for transient error status codes which could be resolved by
    retrying the request. Does not retry ExternalErrors or EndpointErrors
    as those are unlikely to actually be transient.

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted
    """
    retry_config = ctx.caller_info.retry_config
    if ctx.response is not None and (
        ctx.response.status_code in retry_config.transient_error_status_codes
    ):
        try:
            code = ctx.response.json()["code"]
        except (ValueError, KeyError):
            code = ""

        for non_retry_code in ("ExternalError", "EndpointError"):
            if non_retry_code in code:
                return RetryCheckResult.no_decision

        return RetryCheckResult.do_retry

    return RetryCheckResult.no_decision


# Transfer retry checks are the defaults with the transient error one replaced
TRANSFER_DEFAULT_RETRY_CHECKS: tuple[RetryCheck, ...] = tuple(
    check_transfer_transient_error if check is check_transient_error else check
    for check in DEFAULT_RETRY_CHECKS
)



from ..collection import StaticScopeCollection, _urn_scope
from ..representation import Scope


class _AuthScopes(StaticScopeCollection):
    resource_server = "auth.globus.org"

    openid = Scope("openid")
    email = Scope("email")
    profile = Scope("profile")

    manage_projects = _urn_scope(resource_server, "manage_projects")
    view_authentications = _urn_scope(resource_server, "view_authentications")
    view_clients = _urn_scope(resource_server, "view_clients")
    view_clients_and_scopes = _urn_scope(resource_server, "view_clients_and_scopes")
    view_consents = _urn_scope(resource_server, "view_consents")
    view_identities = _urn_scope(resource_server, "view_identities")
    view_identity_set = _urn_scope(resource_server, "view_identity_set")


AuthScopes = _AuthScopes()



from globus_sdk import __version__
from globus_sdk.testing import RegisteredResponse, get_last_request


def test_clientinfo_header_default(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar",
        json={"foo": "bar"},
    ).add()
    res = client.request("GET", "/bar")
    assert res.http_status == 200

    req = get_last_request()
    assert "X-Globus-Client-Info" in req.headers
    assert (
        req.headers["X-Globus-Client-Info"]
        == f"product=python-sdk,version={__version__}"
    )


def test_clientinfo_header_can_be_supressed(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar",
        json={"foo": "bar"},
    ).add()

    # clear the X-Globus-Client-Info header
    client.transport.globus_client_info.clear()

    res = client.request("GET", "/bar")
    assert res.http_status == 200

    req = get_last_request()
    assert "X-Globus-Client-Info" not in req.headers



from __future__ import annotations

import uuid

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "uuid_type",
    (str, uuid.UUID),
)
def test_delete_credential(
    service_client,
    uuid_type: type[str] | type[uuid.UUID],
):
    meta = load_response(service_client.delete_client_credential).metadata

    if uuid_type is str:
        res = service_client.delete_client_credential(
            client_id=meta["client_id"],
            credential_id=meta["credential_id"],
        )
    else:
        res = service_client.delete_client_credential(
            client_id=uuid.UUID(meta["client_id"]),
            credential_id=uuid.UUID(meta["credential_id"]),
        )

    assert res["credential"]["id"] == meta["credential_id"]



import typing as t

import globus_sdk

# test that a SpecificFlowClient allows assignments of scope strings and resource_server
# even though the class-level default is a specialized stub object
flow_id = "foo"
specific_flow_client = globus_sdk.SpecificFlowClient(flow_id)

scopes_object = specific_flow_client.scopes
t.assert_type(scopes_object, globus_sdk.scopes.SpecificFlowScopes)

scope: globus_sdk.Scope = scopes_object.user
x: int = scopes_object.user  # type: ignore[assignment]
resource_server: str = specific_flow_client.scopes.resource_server



from __future__ import annotations

import enum
import typing as t
import uuid

from globus_sdk._internal.remarshal import strseq_iter
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

T = t.TypeVar("T")


class GroupRole(enum.Enum):
    member = "member"
    manager = "manager"
    admin = "admin"


_GROUP_ROLE_T = t.Union[GroupRole, t.Literal["member", "manager", "admin"]]


class GroupMemberVisibility(enum.Enum):
    members = "members"
    managers = "managers"


_GROUP_MEMBER_VISIBILITY_T = t.Union[
    GroupMemberVisibility, t.Literal["members", "managers"]
]


class GroupVisibility(enum.Enum):
    authenticated = "authenticated"
    private = "private"


_GROUP_VISIBILITY_T = t.Union[GroupVisibility, t.Literal["authenticated", "private"]]


class GroupRequiredSignupFields(enum.Enum):
    institution = "institution"
    current_project_name = "current_project_name"
    address = "address"
    city = "city"
    state = "state"
    country = "country"
    address1 = "address1"
    address2 = "address2"
    zip = "zip"
    phone = "phone"
    department = "department"
    field_of_science = "field_of_science"


_GROUP_REQUIRED_SIGNUP_FIELDS_T = t.Union[
    GroupRequiredSignupFields,
    t.Literal[
        "institution",
        "current_project_name",
        "address",
        "city",
        "state",
        "country",
        "address1",
        "address2",
        "zip",
        "phone",
        "department",
        "field_of_science",
    ],
]


def _typename(obj: t.Any) -> str:
    if isinstance(obj, type) and issubclass(obj, enum.Enum):
        return obj.__name__
    return str(obj)


def _fmt_union(obj: t.Any) -> str:
    return " | ".join(_typename(x) for x in t.get_args(obj))


def _docstring_fixer(cls: type[T]) -> type[T]:
    """
    These type aliases are not always rendered correctly by sphinx autodoc.

    Therefore, we state the types explicitly using a doc modifier, so
    that we can reformat them in a sphinx-friendly way.
    """
    if cls.__doc__ is not None:
        cls.__doc__ = cls.__doc__.format(
            _GROUP_ROLE_T=_fmt_union(_GROUP_ROLE_T),
            _GROUP_MEMBER_VISIBILITY_T=_fmt_union(_GROUP_MEMBER_VISIBILITY_T),
            _GROUP_VISIBILITY_T=_fmt_union(_GROUP_VISIBILITY_T),
            _GROUP_REQUIRED_SIGNUP_FIELDS_T=_fmt_union(_GROUP_REQUIRED_SIGNUP_FIELDS_T),
        )
    return cls


class BatchMembershipActions(GlobusPayload):
    """
    An object used to represent a batch action on memberships of a group.
    `Perform actions on group members
    <https://groups.api.globus.org/redoc#operation/group_membership_post_actions_v2_groups__group_id__post>`_.
    """

    def accept_invites(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Accept invites for identities.  The identities must belong to
        the identity set of authenticated user.

        :param identity_ids: The identities for whom to accept invites
        """
        self.setdefault("accept", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def add_members(
        self,
        identity_ids: t.Iterable[uuid.UUID | str],
        *,
        role: _GROUP_ROLE_T = "member",
    ) -> BatchMembershipActions:
        """
        Add a list of identities to a group with the given role.

        :param identity_ids: The identities to add to the group
        :param role: The role for the new group members
        """
        self.setdefault("add", []).extend(
            {"identity_id": identity_id, "role": role}
            for identity_id in strseq_iter(identity_ids)
        )
        return self

    def approve_pending(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Approve a list of identities with pending join requests.

        :param identity_ids: The identities to approve as members of the group
        """
        self.setdefault("approve", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def change_roles(
        self,
        role: _GROUP_ROLE_T,
        identity_ids: t.Iterable[uuid.UUID | str],
    ) -> BatchMembershipActions:
        """
        Assign a new role to a list of identities.

        :param role: The new role to assign.
        :param identity_ids: The identities to assign to the new role.
        """
        self.setdefault("change_role", []).extend(
            {"role": role, "identity_id": identity_id} for identity_id in identity_ids
        )
        return self

    def decline_invites(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Decline an invitation for a given set of identities.

        :param identity_ids: The identities for whom invitations should be declined
        """
        self.setdefault("decline", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def invite_members(
        self,
        identity_ids: t.Iterable[uuid.UUID | str],
        *,
        role: _GROUP_ROLE_T = "member",
    ) -> BatchMembershipActions:
        """
        Invite a list of identities to a group with the given role.

        :param identity_ids: The identities to invite to the group
        :param role: The role for the invited group members
        """
        self.setdefault("invite", []).extend(
            {"identity_id": identity_id, "role": role}
            for identity_id in strseq_iter(identity_ids)
        )
        return self

    def join(self, identity_ids: t.Iterable[uuid.UUID | str]) -> BatchMembershipActions:
        """
        Join a group with the given identities.  The identities must be in the
        authenticated users identity set.

        :param identity_ids: The identities to use to join the group
        """
        self.setdefault("join", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def leave(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Leave a group that one of the identities in the authenticated user's
        identity set is a member of.

        :param identity_ids: The identities to remove from the group
        """
        self.setdefault("leave", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def reject_join_requests(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Reject identities which have requested to join the group.

        :param identity_ids: The identities to reject from the group
        """
        self.setdefault("reject", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def remove_members(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Remove members from a group.  This must be done as an admin or manager
        of the group.

        :param identity_ids: The identities to remove from the group
        """
        self.setdefault("remove", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self

    def request_join(
        self, identity_ids: t.Iterable[uuid.UUID | str]
    ) -> BatchMembershipActions:
        """
        Request to join a group.

        :param identity_ids: The identities to use to request membership in the group
        """
        self.setdefault("request_join", []).extend(
            {"identity_id": identity_id} for identity_id in strseq_iter(identity_ids)
        )
        return self


@_docstring_fixer
class GroupPolicies(GlobusPayload):
    """
    An object used to represent the policy settings of a group.
    This may be used to set or modify group settings.

    See also:
    `API documentation on setting the policies for the group. \
    <https://groups.api.globus.org/redoc#operation/update_policies_v2_groups__group_id__policies_put>`_

    :param is_high_assurance: Whether the group is high assurance or not
    :param group_visibility: The visibility of the group
    :type group_visibility: {_GROUP_VISIBILITY_T}
    :param group_members_visibility: The visibility of the group members
    :type group_members_visibility: {_GROUP_MEMBER_VISIBILITY_T}
    :param join_requests: Whether the group allows join requests or not
    :param signup_fields: The fields required for signup in the group
    :type signup_fields: typing.Iterable[{_GROUP_REQUIRED_SIGNUP_FIELDS_T}]
    :param authentication_assurance_timeout: The session timeout for high assurance
        group policy enforcement
    """

    def __init__(
        self,
        *,
        is_high_assurance: bool,
        group_visibility: _GROUP_VISIBILITY_T,
        group_members_visibility: _GROUP_MEMBER_VISIBILITY_T,
        join_requests: bool,
        signup_fields: t.Iterable[_GROUP_REQUIRED_SIGNUP_FIELDS_T],
        authentication_assurance_timeout: int | None | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["is_high_assurance"] = is_high_assurance
        self["group_visibility"] = group_visibility
        self["group_members_visibility"] = group_members_visibility
        self["join_requests"] = join_requests
        self["signup_fields"] = list(signup_fields)
        self["authentication_assurance_timeout"] = authentication_assurance_timeout



from __future__ import annotations

import enum
import typing as t
import uuid

import requests

from globus_sdk._missing import MISSING, filter_missing


class RequestEncoder:
    """
    A RequestEncoder takes input parameters and outputs a requests.Requests object.

    The default encoder requires that the data is text and is a no-op. It can also be
    referred to as the ``"text"`` encoder.
    """

    def encode(
        self,
        method: str,
        url: str,
        params: dict[str, t.Any] | None,
        data: t.Any,
        headers: dict[str, str],
    ) -> requests.Request:
        if not isinstance(data, (str, bytes)):
            raise TypeError(
                "Cannot encode non-text in a text request. "
                "Either manually encode the data or use `encoding=form|json` to "
                "correctly format this data."
            )
        return requests.Request(
            method,
            url,
            data=self._prepare_data(data),
            params=self._prepare_params(params),
            headers=self._prepare_headers(headers),
        )

    def _format_primitive(self, value: t.Any) -> t.Any:
        """
        Transformations for primitive values (e.g. stringifiable items) for query
        params, headers, and body elements.

        Transforms data as follows:

            x: UUID -> str(x)
            x: Enum -> x.value
            x: _    -> x
        """
        if isinstance(value, uuid.UUID):
            return str(value)
        if isinstance(value, enum.Enum):
            return value.value
        return value

    def _prepare_params(
        self, params: dict[str, t.Any] | None
    ) -> dict[str, t.Any] | None:
        """
        Prepare the query params for a request.

        Filters out MISSING and formats primitives.
        """
        if params is None:
            return None
        return filter_missing({k: self._format_primitive(v) for k, v in params.items()})

    def _prepare_headers(
        self, headers: dict[str, t.Any] | None
    ) -> dict[str, t.Any] | None:
        """
        Prepare the headers for a request.

        Filters out MISSING and formats primitives.
        """
        if headers is None:
            return None
        return filter_missing(
            {k: self._format_primitive(v) for k, v in headers.items()}
        )

    def _prepare_data(self, data: t.Any) -> t.Any:
        """
        Prepare the data (body) for a request.

        If the body is a dict, list, or tuple, it will be recursively processed to
        filter out MISSING and format primitives.

        Otherwise, it is returned as-is.
        """
        if isinstance(data, dict):
            return filter_missing({k: self._prepare_data(v) for k, v in data.items()})
        elif isinstance(data, (list, tuple)):
            return [self._prepare_data(x) for x in data if x is not MISSING]
        else:
            return self._format_primitive(data)


class JSONRequestEncoder(RequestEncoder):
    """
    This encoder prepares the data as JSON. It also ensures that content-type is set, so
    that APIs requiring a content-type of "application/json" are able to read the data.
    """

    def encode(
        self,
        method: str,
        url: str,
        params: dict[str, t.Any] | None,
        data: t.Any,
        headers: dict[str, str],
    ) -> requests.Request:
        if data is not None:
            headers = {"Content-Type": "application/json", **headers}
        return requests.Request(
            method,
            url,
            json=self._prepare_data(data),
            params=self._prepare_params(params),
            headers=self._prepare_headers(headers),
        )


class FormRequestEncoder(RequestEncoder):
    """
    This encoder formats data as a form-encoded body. It requires that the input data is
    a dict -- any other datatype will result in errors.
    """

    def encode(
        self,
        method: str,
        url: str,
        params: dict[str, t.Any] | None,
        data: t.Any,
        headers: dict[str, str],
    ) -> requests.Request:
        if not isinstance(data, dict):
            raise TypeError("FormRequestEncoder cannot encode non-dict data")
        return requests.Request(
            method,
            url,
            data=self._prepare_data(data),
            params=self._prepare_params(params),
            headers=self._prepare_headers(headers),
        )



from __future__ import annotations

import typing as t

from globus_sdk._missing import MISSING

VersionTuple = t.Tuple[int, int, int]

DatatypeCallback = t.Callable[["DocumentWithInducedDatatype"], t.Optional[VersionTuple]]


class DocumentWithInducedDatatype(t.Protocol):
    DATATYPE_BASE: str
    DATATYPE_VERSION_IMPLICATIONS: dict[str, VersionTuple]
    DATATYPE_VERSION_CALLBACKS: tuple[DatatypeCallback, ...]

    def __contains__(self, key: str) -> bool:  # pragma: no cover
        ...

    def __setitem__(self, key: str, value: t.Any) -> None:  # pragma: no cover
        ...

    def __getitem__(self, key: str) -> t.Any:  # pragma: no cover
        ...


def deduce_datatype_version(obj: DocumentWithInducedDatatype) -> str:
    max_deduced_version = (1, 0, 0)
    for fieldname, version in obj.DATATYPE_VERSION_IMPLICATIONS.items():
        if fieldname not in obj or obj[fieldname] is MISSING:
            continue
        if version > max_deduced_version:
            max_deduced_version = version
    for callback in obj.DATATYPE_VERSION_CALLBACKS:
        opt_version = callback(obj)
        if opt_version is not None and opt_version > max_deduced_version:
            max_deduced_version = opt_version
    return ".".join(str(x) for x in max_deduced_version)


def ensure_datatype(obj: DocumentWithInducedDatatype) -> None:
    if "DATA_TYPE" not in obj or obj["DATA_TYPE"] is MISSING:
        obj["DATA_TYPE"] = f"{obj.DATATYPE_BASE}#{deduce_datatype_version(obj)}"



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID, SUBSCRIPTION_ID

OTHER_SUBSCRIPTION_ID = str(uuid.UUID(int=10))

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        json={
            "DATA_TYPE": "result",
            "code": "Updated",
            "message": "Endpoint updated successfully",
            "request_id": "dWTZZe17L",
            "resource": f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        },
    ),
    not_found=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        status=404,
        json={
            "code": "EndpointNotFound",
            "message": f"No such endpoint '{ENDPOINT_ID}'",
            "request_id": "BHI2BHt8N",
            "resource": f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        },
    ),
    multi_subscriber_cannot_use_default=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        status=400,
        json={
            "code": "BadRequest",
            "message": (
                "Please specify the subscription ID to use. "
                "You currently have access to: "
                f"{SUBSCRIPTION_ID}, {OTHER_SUBSCRIPTION_ID}"
            ),
            "request_id": "H1dFNg6QB",
            "resource": f"/v0.10/endpoint/{ENDPOINT_ID}/subscription",
        },
        metadata={
            "endpoint_id": ENDPOINT_ID,
            "subscription_ids": [SUBSCRIPTION_ID, OTHER_SUBSCRIPTION_ID],
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_ID, TWO_HOP_TRANSFER_RUN

_request_params = {
    "body": TWO_HOP_TRANSFER_RUN["details"]["details"]["input"],
    "tags": TWO_HOP_TRANSFER_RUN["tags"],
    "label": TWO_HOP_TRANSFER_RUN["label"],
    "run_monitors": TWO_HOP_TRANSFER_RUN["run_monitors"],
    "run_managers": TWO_HOP_TRANSFER_RUN["run_managers"],
}
RESPONSES = ResponseSet(
    metadata={"flow_id": TWO_HOP_TRANSFER_FLOW_ID, "request_params": _request_params},
    default=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/run",
        json=TWO_HOP_TRANSFER_RUN,
    ),
    missing_scope_error=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/run",
        status=403,
        json={
            "error": {
                "code": "MISSING_SCOPE",
                "detail": (
                    "This action requires the following scope: frobulate[demuddle]"
                ),
            }
        },
    ),
)



import pytest

from globus_sdk.token_storage.legacy import SQLiteAdapter


@pytest.fixture
def db_file(tmp_path):
    return tmp_path / "test.db"


MEMORY_DBNAME = ":memory:"


@pytest.fixture
def adapters_to_close():
    data = set()
    yield data
    for x in data:
        x.close()


@pytest.fixture
def make_adapter(adapters_to_close):
    def func(*args, **kwargs):
        ret = SQLiteAdapter(*args, **kwargs)
        adapters_to_close.add(ret)
        return ret

    return func


@pytest.mark.parametrize(
    "success, use_file, kwargs",
    [
        (False, False, {}),
        (False, False, {"namespace": "foo"}),
        (True, False, {"dbname": MEMORY_DBNAME}),
        (True, False, {"dbname": MEMORY_DBNAME, "namespace": "foo"}),
        (True, True, {}),
        (True, True, {"namespace": "foo"}),
        (False, True, {"dbname": MEMORY_DBNAME}),
        (False, True, {"dbname": MEMORY_DBNAME, "namespace": "foo"}),
    ],
)
def test_constructor(success, use_file, kwargs, db_file, make_adapter):
    if success:
        if use_file:
            make_adapter(db_file, **kwargs)
        else:
            make_adapter(**kwargs)
    else:
        with pytest.raises(TypeError):
            if use_file:
                make_adapter(db_file, **kwargs)
            else:
                make_adapter(**kwargs)


def test_store_and_retrieve_simple_config(make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    store_val = {"val1": True, "val2": None, "val3": 1.4}
    adapter.store_config("myconf", store_val)
    read_val = adapter.read_config("myconf")
    assert read_val == store_val
    assert read_val is not store_val


def test_store_and_retrieve(mock_response, make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    adapter.store(mock_response)

    data = adapter.get_by_resource_server()
    assert data == mock_response.by_resource_server


def test_on_refresh_and_retrieve(mock_response, make_adapter):
    """just confirm that the aliasing of these functions does not change anything"""
    adapter = make_adapter(MEMORY_DBNAME)
    adapter.on_refresh(mock_response)

    data = adapter.get_by_resource_server()
    assert data == mock_response.by_resource_server


def test_multiple_adapters_store_and_retrieve(mock_response, db_file, make_adapter):
    adapter1 = make_adapter(db_file)
    adapter2 = make_adapter(db_file)
    adapter1.store(mock_response)

    data = adapter2.get_by_resource_server()
    assert data == mock_response.by_resource_server


def test_multiple_adapters_store_and_retrieve_different_namespaces(
    mock_response, db_file, make_adapter
):
    adapter1 = make_adapter(db_file, namespace="foo")
    adapter2 = make_adapter(db_file, namespace="bar")
    adapter1.store(mock_response)

    data = adapter2.get_by_resource_server()
    assert data == {}


def test_load_missing_config_data(make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    assert adapter.read_config("foo") is None


def test_load_missing_token_data(make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    assert adapter.get_by_resource_server() == {}
    assert adapter.get_token_data("resource_server_1") is None


def test_remove_tokens(mock_response, make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    adapter.store(mock_response)

    removed = adapter.remove_tokens_for_resource_server("resource_server_1")
    assert removed
    data = adapter.get_by_resource_server()
    assert data == {
        "resource_server_2": mock_response.by_resource_server["resource_server_2"]
    }

    removed = adapter.remove_tokens_for_resource_server("resource_server_1")
    assert not removed


def test_remove_config(make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    store_val = {"val1": True, "val2": None, "val3": 1.4}
    adapter.store_config("myconf", store_val)
    adapter.store_config("myconf2", store_val)
    removed = adapter.remove_config("myconf")
    assert removed
    read_val = adapter.read_config("myconf")
    assert read_val is None
    read_val = adapter.read_config("myconf2")
    assert read_val == store_val

    removed = adapter.remove_config("myconf")
    assert not removed


def test_store_and_refresh(mock_response, mock_refresh_response, make_adapter):
    adapter = make_adapter(MEMORY_DBNAME)
    adapter.store(mock_response)

    # rs1 and rs2 data was stored correctly
    data = adapter.get_token_data("resource_server_1")
    assert data["access_token"] == "access_token_1"
    data = adapter.get_token_data("resource_server_2")
    assert data["access_token"] == "access_token_2"

    # "refresh" happens, this should change rs2 but not rs1
    adapter.store(mock_refresh_response)
    data = adapter.get_token_data("resource_server_1")
    assert data["access_token"] == "access_token_1"
    data = adapter.get_token_data("resource_server_2")
    assert data["access_token"] == "access_token_2_refreshed"


def test_iter_namespaces(mock_response, db_file, make_adapter):
    foo_adapter = make_adapter(db_file, namespace="foo")
    bar_adapter = make_adapter(db_file, namespace="bar")
    baz_adapter = make_adapter(db_file, namespace="baz")

    for adapter in [foo_adapter, bar_adapter, baz_adapter]:
        assert list(adapter.iter_namespaces()) == []
        assert list(adapter.iter_namespaces(include_config_namespaces=True)) == []

    foo_adapter.store(mock_response)

    for adapter in [foo_adapter, bar_adapter, baz_adapter]:
        assert list(adapter.iter_namespaces()) == ["foo"]
        assert list(adapter.iter_namespaces(include_config_namespaces=True)) == ["foo"]

    bar_adapter.store(mock_response)

    for adapter in [foo_adapter, bar_adapter, baz_adapter]:
        assert set(adapter.iter_namespaces()) == {"foo", "bar"}
        assert set(adapter.iter_namespaces(include_config_namespaces=True)) == {
            "foo",
            "bar",
        }

    baz_adapter.store_config("some_conf", {})

    for adapter in [foo_adapter, bar_adapter, baz_adapter]:
        assert set(adapter.iter_namespaces()) == {"foo", "bar"}
        assert set(adapter.iter_namespaces(include_config_namespaces=True)) == {
            "foo",
            "bar",
            "baz",
        }



from ._errors import ConsentParseError, ConsentTreeConstructionError
from ._model import Consent, ConsentForest, ConsentTree

__all__ = (
    "Consent",
    "ConsentTree",
    "ConsentForest",
    "ConsentParseError",
    "ConsentTreeConstructionError",
)



import pytest


def test_paginated_usage_requires_an_argument(
    sphinxext, docutils_runner, register_temporary_directive
):
    register_temporary_directive("paginatedusage", sphinxext.directives.PaginatedUsage)

    with pytest.raises(Exception, match=r"1 argument\(s\) required, 0 supplied\."):
        docutils_runner.to_etree(".. paginatedusage::")


def test_paginated_usage_simple(sphinx_runner):
    etree = sphinx_runner.to_etree(".. paginatedusage:: foo")
    assert etree.tag == "document"
    paragraph = etree.find("paragraph")
    assert paragraph is not None
    text_parts = list(paragraph.itertext())
    assert text_parts[0].startswith("This method supports paginated access.")

    code_block = etree.find("literal_block")
    assert code_block is not None
    assert code_block.get("language") == "default"
    assert code_block.text == "client.paginated.foo(...)"



from __future__ import annotations

import functools
import inspect
import textwrap
import types
import typing as t
from pydoc import locate


def locate_class(classname: str) -> type:
    cls = locate(classname)
    if not inspect.isclass(cls):
        raise RuntimeError(
            f"uh-oh, {classname} is not a class name? type(classname)={type(cls)}"
        )
    return cls


def classname2methods(
    classname: str, include_methods: t.Sequence[str]
) -> list[tuple[str, types.FunctionType]]:
    """Resolve a class name to a list of (public) method names + function objects.
    Takes a classname and a list of method names to avoid filtering out.

    :param classname: The name to resolve to a class
    :param include_methods: A list or tuple of method names which would normally be
        excluded as private, but which should be included in the result
    """
    cls = locate_class(classname)

    # get methods of the object as [(name, <unbound method>), ...]
    methods = inspect.getmembers(cls, predicate=inspect.isfunction)

    def methodname_is_good(m: str) -> bool:
        if m in include_methods:
            return True
        # filter out dunder-methods and `_private` methods
        if m.startswith("_"):
            return False
        # filter out any inherited methods which are not overloaded
        if m not in cls.__dict__:
            return False
        return True

    return [(name, value) for name, value in methods if methodname_is_good(name)]


def derive_doc_url_base(service: str | None) -> str:
    if service is None:
        return "https://docs.globus.org/api"
    elif service == "groups":
        return "https://groups.api.globus.org/redoc#operation"
    elif service == "gcs":
        return "https://docs.globus.org/globus-connect-server/v5/api"
    elif service == "flows":
        return "https://globusonline.github.io/globus-flows#tag"
    elif service == "compute":
        return "https://compute.api.globus.org/redoc#tag"
    else:
        raise ValueError(f"Unsupported extdoclink service '{service}'")


@functools.cache
def read_sphinx_params(docstring: str) -> tuple[str, ...]:
    """
    Given a docstring, extract the `:param:` declarations into a tuple of strings.

    :param docstring: The ``__doc__`` to parse, as it appeared on the original object

    Params start with `:param ...` and end
    - at the end of the string
    - at the next param
    - when a non-indented, non-param line is found

    Whitespace lines within a param doc are supported.

    All produced param strings are dedented.
    """
    docstring = textwrap.dedent(docstring)

    result: list[str] = []
    current: list[str] = []
    for line in docstring.splitlines():
        if not current:
            if line.startswith(":param"):
                current = [line]
            else:
                continue
        else:
            # a new param -- flush the current one and restart
            if line.startswith(":param"):
                result.append("\n".join(current).strip())
                current = [line]
            # a continuation line for the current param (indented)
            # or a blank line -- it *could* be a continuation of param doc (include it)
            elif line != line.lstrip() or not line:
                current.append(line)
            # otherwise this is a populated line, not indented, and without a `:param`
            # start -- stop looking for more param doc
            else:
                break
    if current:
        result.append("\n".join(current).strip())

    return tuple(result)



import pytest

import globus_sdk


@pytest.fixture
def login_client():
    client = globus_sdk.AuthLoginClient()
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def service_client():
    client = globus_sdk.AuthClient()
    with client.retry_config.tune(max_retries=0):
        yield client



from __future__ import annotations

import typing as t

from .base import PageT, Paginator


class MarkerPaginator(Paginator[PageT]):
    """
    A paginator which uses `has_next_page` and `marker` from payloads, sets the `marker`
    query param to page.

    This is the default method for GCS pagination, so it's very simple.
    """

    _REQUIRES_METHOD_KWARGS = ("marker",)

    def __init__(
        self,
        method: t.Callable[..., t.Any],
        *,
        items_key: str | None = None,
        marker_key: str = "marker",
        client_args: tuple[t.Any, ...],
        client_kwargs: dict[str, t.Any],
    ) -> None:
        super().__init__(
            method,
            items_key=items_key,
            client_args=client_args,
            client_kwargs=client_kwargs,
        )
        self.marker: str | None = None
        self.marker_key = marker_key

    def _check_has_next_page(self, page: dict[str, t.Any]) -> bool:
        return bool(page.get("has_next_page", False))

    def pages(self) -> t.Iterator[PageT]:
        has_next_page = True
        while has_next_page:
            if self.marker:
                self.client_kwargs["marker"] = self.marker
            current_page = self.method(*self.client_args, **self.client_kwargs)
            yield current_page
            self.marker = current_page.get(self.marker_key)
            has_next_page = self._check_has_next_page(current_page)


class NullableMarkerPaginator(MarkerPaginator[PageT]):
    """
    A paginator which uses a ``marker`` from payloads and sets the ``marker`` query
    param to page.

    Unlike the base MarkerPaginator, it checks for a null marker to indicate an end to
    pagination. (vs an explicit has_next_page key)
    """

    def _check_has_next_page(self, page: dict[str, t.Any]) -> bool:
        return page.get(self.marker_key) is not None



from globus_sdk.testing import load_response


def test_get_group_by_subscription_id(groups_client):
    meta = load_response(groups_client.get_group_by_subscription_id).metadata

    res = groups_client.get_group_by_subscription_id(meta["subscription_id"])
    assert res.http_status == 200
    assert res["group_id"] == meta["group_id"]
    assert "description" not in res
    assert "name" not in res


def test_two_step_get_group_by_subscription_id(groups_client):
    meta = load_response(groups_client.get_group_by_subscription_id).metadata
    load_response(groups_client.get_group, case="subscription").metadata

    res = groups_client.get_group_by_subscription_id(meta["subscription_id"])
    assert res.http_status == 200
    assert res["group_id"] == meta["group_id"]
    assert "description" not in res
    assert "name" not in res

    res2 = groups_client.get_group(meta["group_id"])
    assert res2.http_status == 200
    assert res2["id"] == meta["group_id"]
    assert "name" in res2



from __future__ import annotations

import inspect
import sys
import uuid

import pytest

from globus_sdk import ConnectorTable, GlobusConnectServerConnector


@pytest.mark.parametrize("connector_data", ConnectorTable._connectors)
def test_lookup_by_attribute(connector_data):
    attrname, connector_name, _ = connector_data

    connector = getattr(ConnectorTable, attrname)
    assert connector.name == connector_name


@pytest.mark.parametrize("connector_data", ConnectorTable._connectors)
@pytest.mark.parametrize("as_uuid", (True, False))
def test_lookup_by_id(connector_data, as_uuid):
    _, connector_name, connector_id = connector_data

    if as_uuid:
        connector_id = uuid.UUID(connector_id)

    connector = ConnectorTable.lookup(connector_id)
    assert connector.name == connector_name


@pytest.mark.parametrize("connector_data", ConnectorTable._connectors)
def test_lookup_by_name(connector_data):
    _, connector_name, connector_id = connector_data

    connector = ConnectorTable.lookup(connector_name)
    assert connector.connector_id == connector_id


@pytest.mark.parametrize(
    "lookup_name, expect_name",
    (
        ("Google Drive", "Google Drive"),
        ("google drive", "Google Drive"),
        ("google_drive", "Google Drive"),
        ("google-drive", "Google Drive"),
        ("google-----drive", "Google Drive"),
        ("google-_-drive", "Google Drive"),  # moody
        ("  google_-drIVE", "Google Drive"),
        ("google_-drIVE    ", "Google Drive"),
        ("   GOOGLE DRIVE    ", "Google Drive"),
    ),
)
def test_lookup_by_name_normalization(lookup_name, expect_name):
    connector = ConnectorTable.lookup(lookup_name)
    assert connector.name == expect_name


@pytest.mark.parametrize("name", [c.name for c in ConnectorTable.all_connectors()])
def test_all_connector_names_map_to_attributes(name):
    connector = ConnectorTable.lookup(name)
    assert connector is not None
    name = name.replace(" ", "_").upper()
    assert getattr(ConnectorTable, name) == connector


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="inspect.get_annotations added in 3.10"
)
def test_all_connector_attributes_are_assigned():
    # build a list of attribute names annotated with
    # `t.ClassVar[GlobusConnectServerConnector]`
    annotated_attributes = []
    for attribute, annotation in inspect.get_annotations(ConnectorTable).items():
        # get_annotations does not interpret string-ized annotations by default, so we
        # receive the relevant values as strings, making comparison simple
        if annotation != "t.ClassVar[GlobusConnectServerConnector]":
            continue
        annotated_attributes.append(attribute)

    # confirm that we got the right number of annotated items
    assert len(annotated_attributes) == len(ConnectorTable._connectors)
    # now confirm that all of these are assigned values
    for attribute in annotated_attributes:
        instance = getattr(ConnectorTable, attribute)
        assert isinstance(instance, GlobusConnectServerConnector)


def test_table_can_be_extended_with_simple_item():
    # don't think too hard about the ethical implications of transporter clones
    connector_name = "Star Trek Transporter"
    connector_id = uuid.uuid1()

    ExtendedTable = ConnectorTable.extend(
        connector_name=connector_name, connector_id=connector_id
    )

    # we get a new subclass named 'ExtendedConnectorTable'
    assert issubclass(ExtendedTable, ConnectorTable)
    assert ExtendedTable.__name__ == "ExtendedConnectorTable"

    # access via name, attribute, and ID all resolve to the same object
    data_object = ExtendedTable.lookup(connector_id)
    assert isinstance(data_object, GlobusConnectServerConnector)
    assert ExtendedTable.STAR_TREK_TRANSPORTER is data_object
    assert ExtendedTable.lookup(connector_name) is data_object


def test_table_extended_twice():
    connector_id1 = uuid.uuid1()
    connector_id2 = uuid.uuid1()

    ExtendedTable1 = ConnectorTable.extend(
        connector_name="Star Trek Transporter", connector_id=connector_id1
    )
    ExtendedTable2 = ExtendedTable1.extend(
        connector_name="Battlestar Galactica FTL", connector_id=connector_id2
    )

    # we get new subclasses named 'ExtendedConnectorTable'
    assert issubclass(ExtendedTable1, ConnectorTable)
    assert ExtendedTable1.__name__ == "ExtendedConnectorTable"
    assert issubclass(ExtendedTable2, ExtendedTable1)
    assert ExtendedTable2.__name__ == "ExtendedConnectorTable"

    # both tables get the same object for connector1
    # only table2 has connector2
    data_object = ExtendedTable1.lookup(connector_id1)
    assert isinstance(data_object, GlobusConnectServerConnector)
    assert ExtendedTable2.lookup(connector_id1) is data_object

    assert ExtendedTable1.lookup(connector_id2) is None
    data_object2 = ExtendedTable2.lookup(connector_id2)
    assert isinstance(data_object2, GlobusConnectServerConnector)



# validate that GlobusAPIError and GlobusHTTPResponse both satisfy the SDK's
# ResponseLike protocol

import sys
import typing as t

import requests

from globus_sdk import GlobusAPIError, GlobusHTTPResponse
from globus_sdk._internal.type_definitions import ResponseLike

if sys.version_info < (3, 11):
    from typing_extensions import assert_type
else:
    from typing import assert_type

sample_response = requests.Response()


def get_length_of_response(r: ResponseLike) -> int:
    return len(r.binary_content)


# confirm this test function is well-declared
get_length_of_response("foo")  # type: ignore[arg-type]


# check an error object
err = GlobusAPIError(sample_response)
assert_type(err.http_status, int)
assert_type(err.http_reason, str)
assert_type(err.headers, t.Mapping[str, str])
assert_type(err.content_type, str | None)
assert_type(err.text, str)
assert_type(err.binary_content, bytes)
assert_type(get_length_of_response(err), int)


# check an HTTP response object
resp = GlobusHTTPResponse(sample_response)
assert_type(resp.http_status, int)
assert_type(resp.http_reason, str)
assert_type(resp.headers, t.Mapping[str, str])
assert_type(resp.content_type, str | None)
assert_type(resp.text, str)
assert_type(resp.binary_content, bytes)
assert_type(get_length_of_response(resp), int)



import globus_sdk

# this is the tutorial client ID
# replace this string with your ID for production use
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# create your app
my_app = globus_sdk.UserApp("my-user-app", client_id=CLIENT_ID)

# create a client with your app
groups_client = globus_sdk.GroupsClient(app=my_app)

# call out to the Groups service to get a listing
my_groups = groups_client.get_my_groups()

# print in CSV format
print("ID,Name,Roles")
for group in my_groups:
    roles = "|".join({m["role"] for m in group["my_memberships"]})
    print(",".join([group["id"], f'"{group["name"]}"', roles]))



import json
import typing as t

from docutils.parsers.rst import directives

from globus_sdk.testing import ResponseList

from .add_content_directive import AddContentDirective


class ExpandTestingFixture(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        "case": directives.unchanged,
    }

    def gen_rst(self) -> t.Iterator[str]:
        from globus_sdk.testing import get_response_set

        response_set_name = self.arguments[0]
        casename = "default"
        if "case" in self.options:
            casename = self.options["case"].strip()
        response_set = get_response_set(response_set_name)
        response = response_set.lookup(casename)
        if isinstance(response, ResponseList):
            # If the default responses is a list of responses, use the first one
            response = response.responses[0]
        if response.json is not None:
            yield ".. code-block:: json"
            yield ""
            output_lines = json.dumps(
                response.json, indent=2, separators=(",", ": ")
            ).split("\n")
            for line in output_lines:
                yield f"    {line}"
        elif response.body is not None:
            yield ".. code-block:: text"
            yield ""
            output_lines = response.body.split("\n")
            for line in output_lines:
                yield f"    {line}"
        else:
            raise RuntimeError(
                "Error loading example content for response "
                f"{response_set_name}:{casename}. Neither JSON nor text body was found."
            )



from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import TASK_DOC, TASK_ID

TASK_BATCH_DOC = {
    "response": "batch",
    "results": {TASK_ID: TASK_DOC},
}

RESPONSES = ResponseSet(
    metadata={"task_id": TASK_ID},
    default=RegisteredResponse(
        service="compute",
        path="/v2/batch_status",
        method="POST",
        json=TASK_BATCH_DOC,
        # Ensure task_ids is a list
        match=[json_params_matcher({"task_ids": [TASK_ID]})],
    ),
)



from __future__ import annotations

import copy
import typing as t
import uuid

from globus_sdk._internal.remarshal import list_map, listify, strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import AbstractGlobusPayload, GlobusPayload

from ._common import DatatypeCallback, ensure_datatype


class StorageGatewayDocument(GlobusPayload):
    """
    Convenience class for constructing a Storage Gateway document
    to use as the `data` parameter to ``create_storage_gateway`` or
    ``update_storage_gateway``

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param display_name: Name of the Storage Gateway
    :param connector_id: UUID of the connector type that this Storage Gateway interacts
        with.
    :param identity_mappings: A list of IdentityMapping objects which are applied to
        user identities to attempt to determine what accounts are available for access.
    :param policies: Connector specific storage policies. It is recommended that
        you use one of the policy helper classes (e.g. `POSIXStoragePolicies`
        if you are using the POSIX connector) to create these.
    :param allowed_domains: List of allowed domains. Users creating credentials or
        collections on this Storage Gateway must have an identity in one of these
        domains.
    :param restrict_paths: Path restrictions within this Storage Gateway. Private.
    :param high_assurance: Flag indicating if the Storage Gateway requires high
        assurance features.
    :param authentication_timeout_mins: Timeout (in minutes) during which a user is
        required to have authenticated in a session to access this storage gateway.
    :param users_allow:  List of connector-specific usernames allowed to access this
        Storage Gateway. Private.
    :param users_deny: List of connector-specific usernames denied access to this
        Storage Gateway. Private.
    :param process_user: Local POSIX user the GridFTP server should run as when
        accessing this Storage Gateway.
    :param load_dsi_module: Name of the DSI module to load by the GridFTP server when
        accessing this Storage Gateway.
    :param require_mfa: Flag indicating that the Storage Gateway requires multi-factor
        authentication. Only usable on high assurance Storage Gateways.

    :param additional_fields: Additional data for inclusion in the Storage Gateway
        document
    """

    DATATYPE_BASE: str = "storage_gateway"
    DATATYPE_VERSION_IMPLICATIONS: dict[str, tuple[int, int, int]] = {
        "require_mfa": (1, 1, 0),
    }
    DATATYPE_VERSION_CALLBACKS: tuple[DatatypeCallback, ...] = ()

    def __init__(
        self,
        DATA_TYPE: str | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        connector_id: uuid.UUID | str | MissingType = MISSING,
        root: str | MissingType = MISSING,
        identity_mappings: t.Iterable[dict[str, t.Any]] | MissingType = MISSING,
        policies: StorageGatewayPolicies | dict[str, t.Any] | MissingType = MISSING,
        allowed_domains: t.Iterable[str] | MissingType = MISSING,
        high_assurance: bool | MissingType = MISSING,
        require_mfa: bool | MissingType = MISSING,
        authentication_timeout_mins: int | MissingType = MISSING,
        users_allow: t.Iterable[str] | MissingType = MISSING,
        users_deny: t.Iterable[str] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["display_name"] = display_name
        self["connector_id"] = connector_id
        self["root"] = root
        self["allowed_domains"] = strseq_listify(allowed_domains)
        self["users_allow"] = strseq_listify(users_allow)
        self["users_deny"] = strseq_listify(users_deny)
        self["high_assurance"] = high_assurance
        self["require_mfa"] = require_mfa
        self["authentication_timeout_mins"] = authentication_timeout_mins
        self["identity_mappings"] = listify(identity_mappings)
        self["policies"] = policies
        self.update(additional_fields or {})
        ensure_datatype(self)


class StorageGatewayPolicies(AbstractGlobusPayload):
    """
    This is the abstract base type for Storage Policies documents to use as the
    ``policies`` parameter when creating a StorageGatewayDocument.

    Several fields on policy documents are marked as ``Private``. This means that they
    are not visible except to admins and owners of the storage gateway.
    """


class POSIXStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a POSIX Storage Policy document
    to use as the ``policies`` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param groups_allow: List of POSIX group IDs allowed to access this Storage
        Gateway. Private.
    :param groups_deny: List of POSIX group IDs denied access this Storage
        Gateway. Private.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_storage_policies#1.0.0",
        groups_allow: t.Iterable[str] | MissingType = MISSING,
        groups_deny: t.Iterable[str] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["groups_allow"] = strseq_listify(groups_allow)
        self["groups_deny"] = strseq_listify(groups_deny)
        self.update(additional_fields or {})


class POSIXStagingStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a POSIX Staging Storage Policy document
    to use as the ``policies`` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param groups_allow: List of POSIX group IDs allowed to access this Storage
        Gateway. Private.
    :param groups_deny: List of POSIX group IDs denied access this Storage
        Gateway. Private.
    :param stage_app: Path to the stage app. Private.
    :param environment: A mapping of variable names to values to set in the environment
        when executing the ``stage_app``. Private.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_staging_storage_policies#1.0.0",
        groups_allow: t.Iterable[str] | MissingType = MISSING,
        groups_deny: t.Iterable[str] | MissingType = MISSING,
        stage_app: str | MissingType = MISSING,
        environment: t.Iterable[dict[str, str]] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["stage_app"] = stage_app
        self["groups_allow"] = strseq_listify(groups_allow)
        self["groups_deny"] = strseq_listify(groups_deny)
        # make shallow copies of all the dicts passed
        self["environment"] = list_map(environment, copy.copy)
        self.update(additional_fields or {})


class BlackPearlStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a BlackPearl Storage Policy document
    to use as the ``policies`` parameter when creating a StorageGatewayDocument.

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param s3_endpoint: The URL of the S3 endpoint of the BlackPearl appliance to use to
        access collections on this Storage Gateway.
    :param bp_access_id_file: Path to the file which provides mappings from usernames
        within the configured identity domain to the ID and secret associated with the
        user's BlackPearl account
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "blackpearl_storage_policies#1.0.0",
        s3_endpoint: str | MissingType = MISSING,
        bp_access_id_file: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["s3_endpoint"] = s3_endpoint
        self["bp_access_id_file"] = bp_access_id_file
        self.update(additional_fields or {})


class BoxStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Box Storage Policy document
    to use as the ``policies`` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param enterpriseID: Identifies which Box Enterprise this Storage Gateway is
        authorized to access. Private.
    :param boxAppSettings: Values that the Storage Gateway uses to identify and
        authenticate with the Box API. Private.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "box_storage_policies#1.0.0",
        enterpriseID: str | MissingType = MISSING,
        boxAppSettings: dict[str, t.Any] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["enterpriseID"] = enterpriseID
        self["boxAppSettings"] = boxAppSettings
        self.update(additional_fields or {})


class CephStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Ceph Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param s3_endpoint: URL of the S3 API endpoint
    :param s3_buckets: List of buckets not owned by the collection owner that will be
        shown in the root of collections created at the base of this Storage Gateway.
    :param ceph_admin_key_id: Administrator key id used to authenticate with the ceph
        admin service to obtain user credentials. Private.
    :param ceph_admin_secret_key: Administrator secret key used to authenticate with the
        ceph admin service to obtain user credentials. Private.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "ceph_storage_policies#1.0.0",
        s3_endpoint: str | MissingType = MISSING,
        s3_buckets: t.Iterable[str] | MissingType = MISSING,
        ceph_admin_key_id: str | MissingType = MISSING,
        ceph_admin_secret_key: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["s3_endpoint"] = s3_endpoint
        self["ceph_admin_key_id"] = ceph_admin_key_id
        self["ceph_admin_secret_key"] = ceph_admin_secret_key
        self["s3_buckets"] = strseq_listify(s3_buckets)
        self.update(additional_fields or {})


class GoogleDriveStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Google Drive Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param client_id: Client ID registered with the Google Application console to access
        Google Drive. Private.
    :param secret: Secret created to access access Google Drive with the client_id in
        this policy. Private.
    :param user_api_rate_quota: User API Rate quota associated with this client ID.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "google_drive_storage_policies#1.0.0",
        client_id: str | MissingType = MISSING,
        secret: str | MissingType = MISSING,
        user_api_rate_quota: int | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["client_id"] = client_id
        self["secret"] = secret
        self["user_api_rate_quota"] = user_api_rate_quota
        self.update(additional_fields or {})


class GoogleCloudStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param client_id: Client ID registered with the Google Application console to
        access Google Drive. Private.
    :param secret: Secret created to access access Google Drive with the client_id in
        this policy. Private.
    :param service_account_key: Credentials for use with service account auth, read
        from a Google-provided json file. Private.
    :param buckets: The list of Google Cloud Storage buckets which the Storage
        Gateway is allowed to access, as well as the list of buckets that will be
        shown in root level directory listings. If this list is unset, bucket
        access is unrestricted and all non public credential accessible buckets
        will be shown in root level directory listings. The value is a list of
        bucket names.
    :param projects: The list of Google Cloud Storage projects which the
        Storage Gateway is allowed to access. If this list is unset, project access
        is unrestricted.  The value is a list of project id strings.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "google_cloud_storage_policies#1.0.0",
        client_id: str | MissingType = MISSING,
        secret: str | MissingType = MISSING,
        service_account_key: dict[str, t.Any] | MissingType = MISSING,
        buckets: t.Iterable[str] | MissingType = MISSING,
        projects: t.Iterable[str] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["client_id"] = client_id
        self["secret"] = secret
        self["buckets"] = strseq_listify(buckets)
        self["projects"] = strseq_listify(projects)
        self["service_account_key"] = service_account_key
        self.update(additional_fields or {})


class OneDriveStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a OneDrive Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param client_id: Client ID registered with the MS Application console to
        access OneDrive. Private.
    :param secret: Secret created to access access MS with the client_id in
        this policy. Private.
    :param tenant: MS Tenant ID from which to allow user logins. Private.
    :param user_api_rate_limit: User API Rate limit associated with this client ID.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "onedrive_storage_policies#1.0.0",
        client_id: str | MissingType = MISSING,
        secret: str | MissingType = MISSING,
        tenant: str | MissingType = MISSING,
        user_api_rate_limit: int | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["client_id"] = client_id
        self["secret"] = secret
        self["tenant"] = tenant
        self["user_api_rate_limit"] = user_api_rate_limit
        self.update(additional_fields or {})


class AzureBlobStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Azure Blob Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param client_id: Client ID registered with the MS Application console to
        access Azure Blob Private.
    :param secret: Secret created to access access MS with the client_id in
        this policy. Private.
    :param tenant: MS Tenant ID from which to allow user logins. Private.
    :param account: Azure Storage account. Private.
    :param auth_type: Auth type: user, service_principal or user_service_principal
    :param adls: ADLS support enabled or not. Private.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "azure_blob_storage_policies#1.0.0",
        client_id: str | MissingType = MISSING,
        secret: str | MissingType = MISSING,
        tenant: str | MissingType = MISSING,
        account: str | MissingType = MISSING,
        auth_type: str | MissingType = MISSING,
        adls: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["client_id"] = client_id
        self["secret"] = secret
        self["tenant"] = tenant
        self["account"] = account
        self["auth_type"] = auth_type
        self["adls"] = adls
        self.update(additional_fields or {})


class S3StoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param s3_endpoint: URL of the S3 API endpoint
    :param s3_buckets: List of buckets not owned by the collection owner that
        will be shown in the root of collections created at the base of this
        Storage Gateway.
    :param s3_user_credential_required: Flag indicating if a Globus User must
        register a user credential in order to create a Guest Collection on this
        Storage Gateway.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "s3_storage_policies#1.0.0",
        s3_endpoint: str | MissingType = MISSING,
        s3_buckets: t.Iterable[str] | MissingType = MISSING,
        s3_user_credential_required: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["s3_endpoint"] = s3_endpoint
        self["s3_user_credential_required"] = s3_user_credential_required
        self["s3_buckets"] = strseq_listify(s3_buckets)
        self.update(additional_fields or {})


class ActiveScaleStoragePolicies(S3StoragePolicies):
    """
    The ActiveScale Storage Policy is an alias for the S3 Storage Policy.
    It even uses S3 policy DATA_TYPE.
    """


class IrodsStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing an iRODS Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param irods_environment_file: Path to iRODS environment file on the endpoint
    :param irods_authentication_file: Path to iRODS authentication file on the endpoint
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "irods_storage_policies#1.0.0",
        irods_environment_file: str | MissingType = MISSING,
        irods_authentication_file: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["irods_environment_file"] = irods_environment_file
        self["irods_authentication_file"] = irods_authentication_file
        self.update(additional_fields or {})


class HPSSStoragePolicies(StorageGatewayPolicies):
    """
    Convenience class for constructing a HPSS Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param authentication_mech: Authentication mechanism to use with HPSS.
    :param authenticator: Authentication credentials to use with HPSS.
    :param uda_checksum_support: Flag indicating whether checksums should be
        stored in metadata.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "hpss_storage_policies#1.0.0",
        authentication_mech: str | MissingType = MISSING,
        authenticator: str | MissingType = MISSING,
        uda_checksum_support: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["authentication_mech"] = authentication_mech
        self["authenticator"] = authenticator
        self["uda_checksum_support"] = uda_checksum_support
        self.update(additional_fields or {})



from __future__ import annotations

import typing as t

from globus_sdk._internal.remarshal import strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload
from globus_sdk.services.gcs.data._common import DatatypeCallback, ensure_datatype


class EndpointDocument(GlobusPayload):
    r"""

    :param data_type: Explicitly set the ``DATA_TYPE`` value for this endpoint document.
        Normally ``DATA_TYPE`` is deduced from the provided parameters and should not be
        set. To maximize compatibility with different versions of GCS, only set this
        value when necessary.

    :param contact_email: Email address of the support contact for this endpoint. This
        is visible to end users so that they may contact your organization for support.
    :param contact_info: Other non-email contact information for the endpoint, e.g.
        phone and mailing address. This is visible to end users for support.
    :param department: Department within organization that runs the server(s).
        Searchable. Unicode string, max 1024 characters, no new lines.
    :param description: A description of the endpoint.
    :param display_name: Friendly name for the endpoint, not unique. Unicode string, no
        new lines (\r or \n). Searchable.
    :param info_link: Link to a web page with more information about the endpoint. The
        administrator is responsible for running a website at this URL and verifying
        that it is accepting public connections.
    :param network_use: Control how Globus interacts with this endpoint over the
        network. Allowed values are:

        * ``normal``: (Default) Uses an average level of concurrency and
          parallelism. The levels depend on the number of physical servers in the
          endpoint.
        * ``minimal``: Uses a minimal level of concurrency and parallelism.
        * ``aggressive``: Uses a high level of concurrency and parallelism.
        * ``custom``: Uses custom values of concurrency and parallelism set by the
          endpoint admin. When setting this level, you must also set the
          ``max_concurrency``, ``preferred_concurrency``, ``max_parallelism``, and
          ``preferred_parallelism`` properties.
    :param organization: Organization that runs the server(s). Searchable. Unicode
        string, max 1024 characters, no new lines.
    :param subscription_id: The id of the subscription that is managing this endpoint.
        This may be the special value DEFAULT when using this as input to PATCH or PUT
        to use the caller’s default subscription id.

    :param keywords: List of search keywords for the endpoint. Unicode string, max 1024
        characters total across all strings.

    :param allow_udt: Allow data transfer on this endpoint using the UDT protocol.
    :param public: Flag indicating whether this endpoint is visible to all other Globus
        users. If false, only users which have been granted a role on the endpoint or
        one of its collections, or belong to a domain allowed access to any of its
        storage gateways may view it.

    :param gridftp_control_channel_port: TCP port for the Globus control channel to
        listen on. By default, the control channel is passed through 443 with an ALPN
        header containing the value "ftp".
    :param max_concurrency: Admin-specified value when the ``network_use`` property’s
        value is ``"custom"``; otherwise the preset value for the specified
        ``network_use``.
    :param max_parallelism: Admin-specified value when the ``network_use`` property’s
        value is ``"custom"``; otherwise the preset value for the specified
        ``network_use``.
    :param preferred_concurrency: Admin-specified value when the ``network_use``
        property’s value is ``"custom"``; otherwise the preset value for the specified
        ``network_use``.
    :param preferred_parallelism: Admin-specified value when the ``network_use``
        property’s value is ``"custom"``; otherwise the preset value for the specified
        ``network_use``.
    """

    DATATYPE_BASE = "endpoint"
    DATATYPE_VERSION_IMPLICATIONS: dict[str, tuple[int, int, int]] = {
        "gridftp_control_channel_port": (1, 1, 0),
    }
    DATATYPE_VERSION_CALLBACKS: tuple[DatatypeCallback, ...] = ()

    # Note: The fields below represent the set of mutable endpoint fields in
    #   an Endpoint#1.2.0 document.
    #   https://docs.globus.org/globus-connect-server/v5.4/api/schemas/Endpoint_1_2_0_schema/
    # Read-only fields (e.g. "gcs_manager_url", "endpoint_id", and
    #   "earliest_last_access") are intentionally omitted as this data class is designed
    #   for input construction not response parsing.
    def __init__(
        self,
        *,
        # data type
        data_type: str | MissingType = MISSING,
        # strs
        contact_email: str | MissingType = MISSING,
        contact_info: str | MissingType = MISSING,
        department: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        info_link: str | MissingType = MISSING,
        network_use: (
            t.Literal["normal", "minimal", "aggressive", "custom"] | MissingType
        ) = MISSING,
        organization: str | MissingType = MISSING,
        # nullable strs
        subscription_id: str | None | MissingType = MISSING,
        # str lists
        keywords: t.Iterable[str] | MissingType = MISSING,
        # bools
        allow_udt: bool | MissingType = MISSING,
        public: bool | MissingType = MISSING,
        # ints
        max_concurrency: int | MissingType = MISSING,
        max_parallelism: int | MissingType = MISSING,
        preferred_concurrency: int | MissingType = MISSING,
        preferred_parallelism: int | MissingType = MISSING,
        # nullable ints
        gridftp_control_channel_port: int | None | MissingType = MISSING,
        # additional fields
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = data_type
        self["contact_email"] = contact_email
        self["contact_info"] = contact_info
        self["department"] = department
        self["description"] = description
        self["display_name"] = display_name
        self["info_link"] = info_link
        self["network_use"] = network_use
        self["organization"] = organization
        self["keywords"] = strseq_listify(keywords)
        self["allow_udt"] = allow_udt
        self["public"] = public
        self["max_concurrency"] = max_concurrency
        self["max_parallelism"] = max_parallelism
        self["preferred_concurrency"] = preferred_concurrency
        self["preferred_parallelism"] = preferred_parallelism
        self["subscription_id"] = subscription_id
        self["gridftp_control_channel_port"] = gridftp_control_channel_port

        if additional_fields is not MISSING:
            self.update(additional_fields)
        ensure_datatype(self)



import pytest

import globus_sdk


@pytest.fixture
def auth_client():
    client = globus_sdk.NativeAppAuthClient("dummy_client_id")
    with client.retry_config.tune(max_retries=0):
        yield client



from globus_sdk.testing import get_last_request, load_response


def test_cancel_task(client):
    meta = load_response(client.cancel_task).metadata

    res = client.cancel_task(meta["task_id"])
    assert res.http_status == 200

    req = get_last_request()
    assert req.body is None



import typing as t
import uuid

TIMER_ID = str(uuid.uuid1())

DEST_EP_ID = str(uuid.uuid1())
SOURCE_EP_ID = str(uuid.uuid1())


V2_TRANSFER_TIMER = {
    "body": {
        "DATA": [
            {
                "DATA_TYPE": "transfer_item",
                "destination_path": "/~/dst.txt",
                "source_path": "/share/godata/file1.txt",
            }
        ],
        "DATA_TYPE": "transfer",
        "delete_destination_extra": False,
        "destination_endpoint": DEST_EP_ID,
        "encrypt_data": False,
        "fail_on_quota_errors": False,
        "notify_on_failed": True,
        "notify_on_inactive": True,
        "notify_on_succeeded": True,
        "preserve_timestamp": False,
        "skip_source_errors": False,
        "source_endpoint": SOURCE_EP_ID,
        "store_base_path_info": False,
        "verify_checksum": True,
    },
    "inactive_reason": None,
    "job_id": TIMER_ID,
    "last_ran_at": None,
    "name": "Very Cool Timer",
    "next_run": "2023-10-27T05:00:00+00:00",
    "number_of_errors": 0,
    "number_of_runs": 0,
    "schedule": {
        "type": "recurring",
        "end": {"count": 2},
        "interval_seconds": 604800,
        "start": "2023-10-27T05:00:00+00:00",
    },
    "status": "new",
    "submitted_at": "2023-10-26T20:31:09+00:00",
    "timer_type": "transfer",
}


FLOW_ID = str(uuid.uuid4())
V2_FLOW_TIMER: dict[str, t.Any] = {
    "callback_body": {
        "body": {"input_key": "input_value"},
        "run_managers": [f"urn:globus:auth:identity:{uuid.uuid4()}"],
    },
    "callback_url": f"https://flows.automate.globus.org/flows/{FLOW_ID}/run",
    "inactive_reason": None,
    "interval": None,
    "job_id": str(uuid.uuid4()),
    "last_ran_at": None,
    "n_errors": 0,
    "n_runs": 0,
    "name": "Very Cool Timer",
    "next_run": "2025-10-27T05:00:00+00:00",
    "results": [],
    "schedule": {
        "datetime": "2025-10-27T05:00:00+00:00",
        "type": "once",
    },
    "scope": (
        f"https://auth.globus.org/scopes/{FLOW_ID}"
        f"/flow_{FLOW_ID.replace('-', '_')}_user"
    ),
    "status": "new",
    "stop_after": {
        "date": None,
        "n_runs": 1,
    },
    "submitted_at": "2025-08-01T20:31:09+00:00",
    "timer_type": "flow",
}


# V1 API data

_transfer_data: dict[str, t.Any] = {
    "data": {
        "action_id": "15jfdBESgveZQ",
        "completion_time": "2022-04-01T19:30:05.973261+00:00",
        "creator_id": "urn:globus:auth:identity:5276fa05-eedf-46c5-919f-ad2d0160d1a9",
        "details": {
            "DATA_TYPE": "task",
            "bytes_checksummed": 0,
            "bytes_transferred": 0,
            "canceled_by_admin": None,
            "canceled_by_admin_message": None,
            "command": "API 0.10",
            "completion_time": None,
            "deadline": "2022-04-02T19:30:07+00:00",
            "delete_destination_extra": False,
            "destination_endpoint": "globus#dest_ep",
            "destination_endpoint_display_name": "Some Dest Endpoint",
            "destination_endpoint_id": DEST_EP_ID,
            "directories": 0,
            "effective_bytes_per_second": 0,
            "encrypt_data": False,
            "event_list": [],
            "fail_on_quota_errors": False,
            "fatal_error": None,
            "faults": 0,
            "files": 0,
            "files_skipped": 0,
            "files_transferred": 0,
            "filter_rules": None,
            "history_deleted": False,
            "is_ok": True,
            "is_paused": False,
            "label": "example timer, run 1",
            "nice_status": "Queued",
            "nice_status_details": None,
            "nice_status_expires_in": -1,
            "nice_status_short_description": "Queued",
            "owner_id": "5276fa05-eedf-46c5-919f-ad2d0160d1a9",
            "preserve_timestamp": False,
            "recursive_symlinks": "ignore",
            "request_time": "2022-04-01T19:30:07+00:00",
            "skip_source_errors": True,
            "source_endpoint": "globus#src_ep",
            "source_endpoint_display_name": "Some Source Endpoint",
            "source_endpoint_id": SOURCE_EP_ID,
            "status": "ACTIVE",
            "subtasks_canceled": 0,
            "subtasks_expired": 0,
            "subtasks_failed": 0,
            "subtasks_pending": 1,
            "subtasks_retrying": 0,
            "subtasks_skipped_errors": 0,
            "subtasks_succeeded": 0,
            "subtasks_total": 1,
            "symlinks": 0,
            "sync_level": 3,
            "task_id": "22f0148c-b1f2-11ec-b87e-3912f602f346",
            "type": "TRANSFER",
            "username": "u_kj3pubpo35dmlem7vuwqcygrve",
            "verify_checksum": True,
        },
        "display_status": "ACTIVE",
        "label": None,
        "manage_by": [],
        "monitor_by": [],
        "release_after": "P30D",
        "start_time": "2022-04-01T19:30:05.973232+00:00",
        "status": "ACTIVE",
    },
    "errors": None,
    "status": 202,
    "ran_at": "2022-04-01T19:30:07.103090",
}
V1_TIMER = {
    "name": "example timer",
    "start": "2022-04-01T19:30:00+00:00",
    "stop_after": None,
    "interval": 864000.0,
    "callback_url": "https://actions.automate.globus.org/transfer/transfer/run",
    "callback_body": {
        "body": {
            "label": "example timer",
            "skip_source_errors": True,
            "sync_level": 3,
            "verify_checksum": True,
            "source_endpoint_id": "aa752cea-8222-5bc8-acd9-555b090c0ccb",
            "destination_endpoint_id": "313ce13e-b597-5858-ae13-29e46fea26e6",
            "transfer_items": [
                {
                    "source_path": "/share/godata/file1.txt",
                    "destination_path": "/~/file1.txt",
                    "recursive": False,
                }
            ],
        }
    },
    "inactive_reason": None,
    "scope": None,
    "job_id": TIMER_ID,
    "status": "loaded",
    "submitted_at": "2022-04-01T19:29:55.942546+00:00",
    "last_ran_at": "2022-04-01T19:30:07.103090+00:00",
    "next_run": "2022-04-11T19:30:00+00:00",
    "n_runs": 1,
    "n_errors": 0,
    "results": {"data": [_transfer_data], "page_next": None},
}



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = "60d1160b-f016-40b0-8545-99619865873d"
IDENTITY_IDS = [
    "ae332d86-d274-11e5-b885-b31714a110e9",
    "c699d42e-d274-11e5-bf75-1fc5bf53bb24",
]
ROLE_IDS = ["MDMwMjM5", "MDQ0ODYz"]


RESPONSES = ResponseSet(
    metadata={"index_id": INDEX_ID, "identity_ids": IDENTITY_IDS, "role_ids": ROLE_IDS},
    default=RegisteredResponse(
        service="search",
        path=f"/v1/index/{INDEX_ID}/role_list",
        json={
            "role_list": [
                {
                    "creation_date": "2021-11-09 20:26:45",
                    "id": ROLE_IDS[0],
                    "index_id": INDEX_ID,
                    "principal": "urn:globus:auth:identity:" + IDENTITY_IDS[0],
                    "principal_type": "identity",
                    "role_name": "owner",
                },
                {
                    "creation_date": "2022-01-24 15:33:41",
                    "id": ROLE_IDS[1],
                    "index_id": INDEX_ID,
                    "principal": "urn:globus:auth:identity:" + IDENTITY_IDS[1],
                    "principal_type": "identity",
                    "role_name": "writer",
                },
            ]
        },
    ),
)



import json

import pytest

from globus_sdk import FlowsAPIError
from globus_sdk.testing import get_last_request, load_response


def test_get_run_definition(flows_client):
    """Validate the HTTP method and route used to get the flow definition for a run."""

    run_id = load_response(flows_client.get_run_definition).metadata["run_id"]

    flows_client.get_run_definition(run_id)
    request = get_last_request()
    assert request.method == "GET"
    assert request.url.endswith(f"/runs/{run_id}/definition")


def test_cancel_run(flows_client):
    """Verify that run cancellation requests meet expectations."""

    run_id = load_response(flows_client.cancel_run).metadata["run_id"]

    flows_client.cancel_run(run_id)
    request = get_last_request()
    assert request.method == "POST"
    assert request.url.endswith(f"/runs/{run_id}/cancel")


@pytest.mark.parametrize(
    "values",
    (
        {},
        {"label": "x"},
        {"run_monitors": []},
        {"run_monitors": ["me", "you"]},
        {"run_managers": []},
        {"run_managers": ["me", "you"]},
        {"tags": []},
        {"tags": ["x"]},
    ),
)
def test_update_run(flows_client, values):
    metadata = load_response(flows_client.update_run).metadata

    flows_client.update_run(metadata["run_id"], **values)
    request = get_last_request()
    assert request.method == "PUT"
    assert request.url.endswith(f"/runs/{metadata['run_id']}")
    assert json.loads(request.body) == values

    # Ensure deprecated routes are not used.
    assert f"/flows/{metadata['flow_id']}" not in request.url


def test_update_run_additional_fields(flows_client):
    """*addition_fields* must override all other parameters."""

    metadata = load_response(flows_client.update_run).metadata

    additional_fields = {
        "label": "x",
        "run_monitors": ["x"],
        "run_managers": ["x"],
        "tags": ["x"],
    }

    flows_client.update_run(
        metadata["run_id"],
        label="a",
        run_managers=["a"],
        run_monitors=["a"],
        tags=["a"],
        additional_fields=additional_fields,
    )
    request = get_last_request()
    assert request.method == "PUT"
    assert request.url.endswith(f"/runs/{metadata['run_id']}")
    assert json.loads(request.body) == additional_fields


def test_delete_run_success(flows_client):
    """Verify `.delete_run()` requests match expectations."""

    metadata = load_response(flows_client.delete_run).metadata
    flows_client.delete_run(metadata["run_id"])

    request = get_last_request()
    assert request.method == "POST"
    assert request.url.endswith(f"/runs/{metadata['run_id']}/release")
    # Ensure no deprecated routes are used.
    assert "/flows/" not in request.url


def test_delete_run_conflict(flows_client):
    """Verify the `.delete_run()` HTTP 409 CONFLICT test case matches expectations."""

    metadata = load_response(flows_client.delete_run, case="conflict").metadata

    with pytest.raises(FlowsAPIError) as error:
        flows_client.delete_run(metadata["run_id"])

    assert error.value.http_status == 409



import json
import re
from collections import namedtuple
from unittest import mock

import pytest
import requests
import responses

from globus_sdk.response import ArrayResponse, GlobusHTTPResponse, IterableResponse

_TestResponse = namedtuple("_TestResponse", ("data", "r"))


def _response(data=None, encoding="utf-8", headers=None, status: int = 200):
    r = requests.Response()

    is_json = isinstance(data, (dict, list))

    datastr = json.dumps(data) if is_json else data
    if datastr is not None:
        if isinstance(datastr, str):
            r._content = datastr.encode("utf-8")
            r.encoding = "utf-8"
        else:
            r._content = datastr
            r.encoding = "ISO-8559-1"

    if headers:
        r.headers.update(headers)
    elif is_json:
        r.headers["Content-Type"] = "application/json"

    r.status_code = status

    r.reason = {200: "OK", 404: "Not Found"}.get(status, "Unknown")

    return r


def _mk_json_response(data):
    json_response = _response(data)
    return _TestResponse(data, GlobusHTTPResponse(json_response, client=mock.Mock()))


@pytest.fixture
def dict_response():
    return _mk_json_response({"label1": "value1", "label2": "value2"})


@pytest.fixture
def list_response():
    return _mk_json_response(["value1", "value2", "value3"])


@pytest.fixture
def http_no_content_type_response():
    res = _response()
    assert "Content-Type" not in res.headers
    return _TestResponse(None, GlobusHTTPResponse(res, client=mock.Mock()))


@pytest.fixture
def malformed_http_response():
    malformed_response = _response(b"{", headers={"Content-Type": "application/json"})
    return _TestResponse(
        "{", GlobusHTTPResponse(malformed_response, client=mock.Mock())
    )


@pytest.fixture
def text_http_response():
    text_data = "text data"
    text_response = _response(
        text_data, encoding="utf-8", headers={"Content-Type": "text/plain"}
    )
    return _TestResponse(
        text_data, GlobusHTTPResponse(text_response, client=mock.Mock())
    )


def test_data(
    dict_response,
    list_response,
    malformed_http_response,
    text_http_response,
):
    """
    Gets the data from the GlobusResponses, confirms results
    Gets the data from each HTTPResponse, confirms expected data from json
    and None from malformed or plain text HTTP
    """
    assert dict_response.r.data == dict_response.data
    assert list_response.r.data == list_response.data
    assert malformed_http_response.r.data is None
    assert text_http_response.r.data is None


def test_str(dict_response, list_response):
    """
    Confirms that individual values are seen in stringified responses
    """
    for item in dict_response.data:
        assert item in str(dict_response.r)
    assert "nonexistent" not in str(dict_response.r)

    for item in list_response.data:
        assert item in str(list_response.r)
    assert "nonexistent" not in str(list_response.r)


def test_text_response_repr_and_str_contain_raw_data():
    expect_text = """pu-erh is a distinctive aged tea primarily produced in Yunnan

    depending on the tea used and how it is aged, it can be bright, floral, and fruity
    or it can take on mushroomy, fermented, and malty notes
    """
    raw = _response(
        expect_text, encoding="utf-8", headers={"Content-Type": "text/plain"}
    )
    res = GlobusHTTPResponse(raw, client=mock.Mock())

    assert expect_text in repr(res)
    assert expect_text in str(res)


def test_getitem(dict_response, list_response):
    """
    Confirms that values can be accessed from the GlobusResponse
    """
    # str indexing
    for key in dict_response.data:
        assert dict_response.r[key] == dict_response.data[key]
    # int indexing
    for i in range(len(list_response.data)):
        assert list_response.r[i] == list_response.data[i]
    # slice indexing
    assert list_response.r[:-1] == list_response.data[:-1]


def test_contains(dict_response, list_response, text_http_response):
    """
    Confirms that individual values are seen in the GlobusResponse
    """
    for item in dict_response.data:
        assert item in dict_response.r
    assert "nonexistent" not in dict_response.r

    for item in list_response.data:
        assert item in list_response.r
    assert "nonexistent" not in list_response.r

    assert "foo" not in text_http_response.r


def test_bool(dict_response, list_response):
    assert bool(dict_response) is True
    assert bool(list_response) is True

    empty_dict, empty_list = _mk_json_response({}), _mk_json_response([])
    assert bool(empty_dict.r) is False
    assert bool(empty_list.r) is False

    null = _mk_json_response(None)
    assert bool(null.r) is False


def test_len_array(list_response):
    array = ArrayResponse(list_response.r)
    assert len(array) == len(list_response.data)

    empty_list = _mk_json_response([])
    empty_array = ArrayResponse(empty_list.r)
    assert len(empty_list.data) == 0
    assert len(empty_array) == 0


def test_len_array_bad_data(dict_response):
    null_array = ArrayResponse(_mk_json_response(None).r)
    with pytest.raises(
        TypeError,
        match=re.escape(
            "Cannot take len() on ArrayResponse data when type is 'NoneType'"
        ),
    ):
        len(null_array)

    dict_array = ArrayResponse(dict_response.r)
    with pytest.raises(
        TypeError,
        match=re.escape("Cannot take len() on ArrayResponse data when type is 'dict'"),
    ):
        len(dict_array)


def test_iter_array_bad_data(dict_response):
    null_array = ArrayResponse(_mk_json_response(None).r)
    with pytest.raises(
        TypeError,
        match=re.escape("Cannot iterate on ArrayResponse data when type is 'NoneType'"),
    ):
        list(null_array)

    dict_array = ArrayResponse(dict_response.r)
    with pytest.raises(
        TypeError,
        match=re.escape("Cannot iterate on ArrayResponse data when type is 'dict'"),
    ):
        list(dict_array)


def test_get(dict_response, list_response, text_http_response):
    """
    Gets individual values from dict response, confirms results
    Confirms list response correctly fails as non indexable
    """
    for item in dict_response.data:
        assert dict_response.r.get(item) == dict_response.data.get(item)

    assert list_response.r.get("value1") is None
    assert list_response.r.get("value1", "foo") == "foo"

    assert text_http_response.r.get("foo") is None
    assert text_http_response.r.get("foo", default="bar") == "bar"


def test_text(malformed_http_response, text_http_response):
    """
    Gets the text from each HTTPResponse, confirms expected results
    """
    assert malformed_http_response.r.text == "{"
    assert text_http_response.r.text == text_http_response.data


def test_binary_content_property(malformed_http_response, text_http_response):
    """
    Gets the text from each HTTPResponse, confirms expected results
    """
    assert malformed_http_response.r.binary_content == b"{"
    assert text_http_response.r.binary_content == text_http_response.data.encode(
        "utf-8"
    )


def test_no_content_type_header(http_no_content_type_response):
    """
    Response without a Content-Type HTTP header should be okay
    """
    assert http_no_content_type_response.r.content_type is None


def test_client_required_with_requests_response():
    r = _response({"foo": 1})
    GlobusHTTPResponse(r, client=mock.Mock())  # ok
    with pytest.raises(ValueError):
        GlobusHTTPResponse(r)  # not ok


def test_client_forbidden_when_wrapping():
    r = _response({"foo": 1})
    to_wrap = GlobusHTTPResponse(r, client=mock.Mock())

    GlobusHTTPResponse(to_wrap)  # ok
    with pytest.raises(ValueError):
        GlobusHTTPResponse(to_wrap, client=mock.Mock())  # not ok


def test_value_error_indexing_on_non_json_data():
    r = _response(b"foo: bar, baz: buzz")
    res = GlobusHTTPResponse(r, client=mock.Mock())

    with pytest.raises(ValueError):
        res["foo"]


def test_cannot_construct_base_iterable_response():
    r = _response(b"foo: bar, baz: buzz")
    with pytest.raises(TypeError):
        IterableResponse(r, client=mock.Mock())


def test_iterable_response_using_iter_key():
    class MyIterableResponse(IterableResponse):
        default_iter_key = "default_iter"

    raw = _response({"default_iter": [0, 1], "other_iter": [3, 4]})

    default = MyIterableResponse(raw, client=mock.Mock())
    assert list(default) == [0, 1]

    withkey = MyIterableResponse(raw, client=mock.Mock(), iter_key="other_iter")
    assert list(withkey) == [3, 4]


def test_iterable_response_errors_on_non_dict_data(list_response):
    class MyIterableResponse(IterableResponse):
        default_iter_key = "default_iter"

    list_iterable = MyIterableResponse(list_response.r)
    null_iterable = MyIterableResponse(_mk_json_response(None).r)

    with pytest.raises(
        TypeError,
        match=re.escape("Cannot iterate on IterableResponse data when type is 'list'"),
    ):
        list(list_iterable)

    with pytest.raises(
        TypeError,
        match=re.escape(
            "Cannot iterate on IterableResponse data when type is 'NoneType'"
        ),
    ):
        list(null_iterable)


def test_can_iter_array_response(list_response):
    arr = ArrayResponse(list_response.r)
    # sorted/reversed are just example stdlib functions which use iter
    assert sorted(arr) == sorted(list_response.data)
    assert list(reversed(arr)) == list(reversed(list_response.data))


def test_http_status_code_on_response():
    r1 = _response(status=404)
    assert r1.status_code == 404

    r2 = GlobusHTTPResponse(r1, client=mock.Mock())  # handle a Response object
    assert r2.http_status == 404

    r3 = GlobusHTTPResponse(r2)  # wrap another response
    assert r3.http_status == 404


def test_http_reason_on_response():
    r1 = _response(status=404)
    r2 = GlobusHTTPResponse(r1, client=mock.Mock())  # handle a Response object
    r3 = GlobusHTTPResponse(r2)  # wrap another response
    assert r1.reason == "Not Found"
    assert r2.http_reason == "Not Found"
    assert r3.http_reason == "Not Found"

    r4 = _response(status=200)
    r5 = GlobusHTTPResponse(r4, client=mock.Mock())  # handle a Response object
    r6 = GlobusHTTPResponse(r5)  # wrap another response
    assert r4.reason == "OK"
    assert r5.http_reason == "OK"
    assert r6.http_reason == "OK"


def test_http_headers_from_response():
    r1 = _response(headers={"Content-Length": "5"})
    assert r1.headers["content-length"] == "5"

    r2 = GlobusHTTPResponse(r1, client=mock.Mock())  # handle a Response object
    assert r2.headers["content-length"] == "5"

    r3 = GlobusHTTPResponse(r2)  # wrap another response
    assert r3.headers["content-length"] == "5"


def test_streaming_response_does_not_read_body_on_init():
    # create a streaming response with a trivial body
    responses.add("GET", "https://www.globus.org/", json={})
    requests_response = requests.get("https://www.globus.org/", stream=True)

    # patch the underlying response stream to error -- not because it's important
    # to test the error behavior, but because it's a very clear signal about when/if
    # streaming is forced to evaluate
    with mock.patch.object(
        requests_response.raw, "stream", side_effect=RuntimeError("ohnoez")
    ):
        # no error on init
        sdk_response = GlobusHTTPResponse(requests_response, client=mock.Mock())

        # but accessing data causes a read, and therefore an error from streaming
        with pytest.raises(RuntimeError, match="ohnoez"):
            sdk_response.text



from globus_sdk.scopes import Scope, ScopeParser


def test_base_scope_strings():
    s1 = [Scope("foo"), Scope("bar")]
    s2 = [Scope("foo"), Scope("baz")]
    merged = ScopeParser.merge_scopes(s1, s2)
    assert len(merged) == 3

    str_list = [str(s) for s in merged]
    assert "foo" in str_list
    assert "bar" in str_list
    assert "baz" in str_list


def test_mixed_optional_dependencies():
    s1 = [Scope("foo", optional=True)]
    s2 = [Scope("foo", optional=False)]
    merged = ScopeParser.merge_scopes(s1, s2)
    assert len(merged) == 2

    str_list = [str(s) for s in merged]
    assert "foo" in str_list
    assert "*foo" in str_list


def test_different_dependencies():
    s1 = [Scope("foo").with_dependency(Scope("bar"))]
    s2 = [Scope("foo").with_dependency(Scope("baz"))]
    merged = ScopeParser.merge_scopes(s1, s2)
    assert len(merged) == 1
    assert merged[0].scope_string == "foo"

    dependency_str_list = [str(s) for s in merged[0].dependencies]
    assert len(dependency_str_list) == 2
    assert "bar" in dependency_str_list
    assert "baz" in dependency_str_list


def test_optional_dependencies():
    s1 = [Scope("foo").with_dependency(Scope("bar"))]
    s2 = [Scope("foo").with_dependency(Scope("bar", optional=True))]
    merged = ScopeParser.merge_scopes(s1, s2)
    assert len(merged) == 1
    assert merged[0].scope_string == "foo"

    dependency_str_list = [str(s) for s in merged[0].dependencies]
    assert len(dependency_str_list) == 1
    assert "bar" in dependency_str_list


def test_different_dependencies_on_mixed_optional_base():
    s1 = [Scope("foo").with_dependency(Scope("bar"))]
    s2 = [Scope("foo", optional=True).with_dependency(Scope("baz"))]
    merged = ScopeParser.merge_scopes(s1, s2)
    assert len(merged) == 2

    for scope in merged:
        dependency_str_list = [str(s) for s in scope.dependencies]
        assert len(dependency_str_list) == 2
        assert "bar" in dependency_str_list
        assert "baz" in dependency_str_list



import base64
import os
import shutil
import uuid

import pytest

import globus_sdk
from globus_sdk.testing import load_response

_IS_WINDOWS = os.name == "nt"

ID_ZERO = uuid.UUID(int=0)


def _compute_base32_id(value):
    """
    Create a base32-encoded UUID in the format used by legacy Globus systems to
    represent Identity IDs as unix-friendly usernames.
    GCP CN values still match this format.
    """
    if isinstance(value, str):
        value = uuid.UUID(value)
    encoded_bytes = base64.b32encode(value.bytes)
    result = encoded_bytes.decode()
    result = result.rstrip("=").lower()
    return f"u_{result}"


def _compute_confdir(homedir, alt=False):
    if alt:
        return os.path.join(homedir, "alt-conf-dir/lta")
    if _IS_WINDOWS:
        return os.path.join(homedir, "Globus Connect")
    else:
        return os.path.join(homedir, ".globusonline/lta")


def normalize_config_dir_argument(config_dir):
    return config_dir if _IS_WINDOWS else os.path.dirname(config_dir)


@pytest.fixture
def mocked_confdir(tmp_path):
    confdir = _compute_confdir(tmp_path)
    os.makedirs(confdir)
    return confdir


@pytest.fixture
def mocked_alternate_confdir(tmp_path):
    altconfdir = _compute_confdir(tmp_path, alt=True)
    os.makedirs(altconfdir)
    return altconfdir


@pytest.fixture(autouse=True)
def mocked_homedir(monkeypatch, tmp_path, mocked_confdir, mocked_alternate_confdir):
    def mock_expanduser(path):
        return str(tmp_path / path.replace("~/", ""))

    if _IS_WINDOWS:
        monkeypatch.setitem(os.environ, "LOCALAPPDATA", str(tmp_path))
    else:
        monkeypatch.setattr(os.path, "expanduser", mock_expanduser)


@pytest.fixture
def write_gcp_id_file(mocked_confdir, mocked_alternate_confdir):
    def _func_fixture(epid, alternate=False):
        fpath = os.path.join(
            mocked_alternate_confdir if alternate else mocked_confdir, "client-id.txt"
        )
        with open(fpath, "w") as f:
            f.write(epid)
            f.write("\n")

    return _func_fixture


@pytest.fixture
def write_gridmap(mocked_confdir):
    def _func_fixture(data, alternate=False):
        fpath = os.path.join(
            mocked_alternate_confdir if alternate else mocked_confdir, "gridmap"
        )
        with open(fpath, "w") as f:
            f.write(data)

    return _func_fixture


@pytest.fixture
def local_gcp():
    return globus_sdk.LocalGlobusConnectPersonal()


@pytest.fixture
def auth_client():
    return globus_sdk.AuthClient()


@pytest.mark.skipif(not _IS_WINDOWS, reason="test requires Windows")
def test_localep_localappdata_notset(local_gcp, monkeypatch):
    monkeypatch.delitem(os.environ, "LOCALAPPDATA")
    with pytest.raises(globus_sdk.GlobusSDKUsageError):
        local_gcp.endpoint_id


def test_localep_load_id(local_gcp, write_gcp_id_file):
    assert local_gcp.endpoint_id is None
    write_gcp_id_file("foobar")
    assert local_gcp.endpoint_id == "foobar"
    write_gcp_id_file("xyz")
    assert local_gcp.endpoint_id == "foobar"
    del local_gcp.endpoint_id
    assert local_gcp.endpoint_id == "xyz"


def test_localep_load_id_alternate_conf_dir(
    mocked_alternate_confdir, write_gcp_id_file
):
    gcp = globus_sdk.LocalGlobusConnectPersonal(
        config_dir=normalize_config_dir_argument(mocked_alternate_confdir)
    )
    assert gcp.endpoint_id is None
    write_gcp_id_file("foobar", alternate=True)
    assert gcp.endpoint_id == "foobar"
    write_gcp_id_file("xyz", alternate=True)
    assert gcp.endpoint_id == "foobar"
    del gcp.endpoint_id
    assert gcp.endpoint_id == "xyz"


def test_load_id_no_confdir(local_gcp, mocked_confdir, mocked_alternate_confdir):
    shutil.rmtree(mocked_confdir)
    shutil.rmtree(mocked_alternate_confdir)
    alt_gcp = globus_sdk.LocalGlobusConnectPersonal(config_dir=mocked_alternate_confdir)
    assert local_gcp.endpoint_id is None
    assert alt_gcp.endpoint_id is None


def test_get_owner_info(local_gcp, write_gridmap, auth_client):
    meta = load_response(auth_client.get_identities, case="globusid").metadata
    local_username = meta["short_username"]
    write_gridmap(
        '"/C=US/O=Globus Consortium/OU=Globus Connect User/'
        f'CN={local_username}" {local_username}\n'
    )
    info = local_gcp.get_owner_info()
    assert isinstance(info, globus_sdk.GlobusConnectPersonalOwnerInfo)
    assert info.username == meta["username"]
    assert info.id is None
    assert str(info) == f"GlobusConnectPersonalOwnerInfo(username={info.username})"

    data = local_gcp.get_owner_info(auth_client)
    assert isinstance(data, dict)
    assert data["id"] == meta["id"]


def test_get_owner_info_b32_mode(local_gcp, write_gridmap, auth_client):
    meta = load_response(auth_client.get_identities).metadata
    base32_id = _compute_base32_id(meta["id"])
    local_username = meta["username"].partition("@")[0]
    write_gridmap(
        '"/C=US/O=Globus Consortium/OU=Globus Connect User/'
        f'CN={base32_id}" {local_username}\n'
    )
    info = local_gcp.get_owner_info()
    assert isinstance(info, globus_sdk.GlobusConnectPersonalOwnerInfo)
    assert info.username is None
    assert info.id == meta["id"]

    data = local_gcp.get_owner_info(auth_client)
    assert isinstance(data, dict)
    assert data["id"] == meta["id"]
    assert data["username"] == meta["username"]


# these things are close to the right thing, but each is somehow wrong
@pytest.mark.parametrize(
    "cn",
    [
        # no 'u_'
        _compute_base32_id(ID_ZERO)[2:],
        # short one char
        _compute_base32_id(ID_ZERO)[:-1],
        # invalid b32 char included
        _compute_base32_id(ID_ZERO)[:-1] + "/",
    ],
)
def test_get_owner_info_b32_mode_invalid_data(
    local_gcp, write_gridmap, cn, auth_client
):
    write_gridmap(
        f'"/C=US/O=Globus Consortium/OU=Globus Connect User/CN={cn}" javert\n'
    )
    info = local_gcp.get_owner_info()
    assert isinstance(info, globus_sdk.GlobusConnectPersonalOwnerInfo)
    assert info.username == f"{cn}@globusid.org"


@pytest.mark.parametrize(
    "bad_cn_line",
    [
        '"/C=US/O=Globus Consortium/OU=Globus Connect User/CN=koala"',
        '"/C=US/O=Globus Consortium/OU=Globus Connect User/CN=koala" panda fossa',
        "",
        '"" koala',
    ],
)
def test_get_owner_info_malformed_entry(local_gcp, write_gridmap, bad_cn_line):
    write_gridmap(bad_cn_line + "\n")
    assert local_gcp.get_owner_info() is None


def test_get_owner_info_no_conf(local_gcp):
    assert local_gcp.get_owner_info() is None
    assert local_gcp.get_owner_info(auth_client) is None


def test_get_owner_info_no_confdir(local_gcp, mocked_confdir, auth_client):
    shutil.rmtree(mocked_confdir)
    assert local_gcp.get_owner_info() is None
    assert local_gcp.get_owner_info(auth_client) is None


def test_get_owner_info_multiline_data(local_gcp, write_gridmap, auth_client):
    meta = load_response(auth_client.get_identities, case="globusid").metadata
    local_username = meta["short_username"]
    write_gridmap(
        "\n".join(
            [
                (
                    '"/C=US/O=Globus Consortium/OU=Globus Connect User/'
                    f'CN={local_username}{x}" {local_username}{x}'
                )
                for x in ["", "2", "3"]
            ]
        )
        + "\n"
    )
    info = local_gcp.get_owner_info()
    assert isinstance(info, globus_sdk.GlobusConnectPersonalOwnerInfo)
    assert info.username == meta["username"]

    data = local_gcp.get_owner_info(auth_client)
    assert isinstance(data, dict)
    assert data["id"] == meta["id"]


def test_get_owner_info_no_auth_data(local_gcp, write_gridmap, auth_client):
    load_response(auth_client.get_identities, case="empty")
    write_gridmap(
        '"/C=US/O=Globus Consortium/OU=Globus Connect User/CN=azathoth" azathoth\n'
    )
    info = local_gcp.get_owner_info()
    assert isinstance(info, globus_sdk.GlobusConnectPersonalOwnerInfo)
    assert info.username == "azathoth@globusid.org"

    data = local_gcp.get_owner_info(auth_client)
    assert data is None


def test_get_owner_info_gridmap_permission_denied(local_gcp, mocked_confdir):
    fpath = os.path.join(mocked_confdir, "gridmap")
    if not _IS_WINDOWS:
        with open(fpath, "w"):  # "touch"
            pass
        os.chmod(fpath, 0o000)
    else:
        # on windows, trying to read a directory gets a permission error
        # this is just an easy way for tests to simulate bad permissions
        os.makedirs(fpath)

    with pytest.raises(PermissionError):
        local_gcp.get_owner_info()


def test_get_endpoint_id_permission_denied(local_gcp, mocked_confdir):
    fpath = os.path.join(mocked_confdir, "client-id.txt")
    if not _IS_WINDOWS:
        with open(fpath, "w"):  # "touch"
            pass
        os.chmod(fpath, 0o000)
    else:
        # on windows, trying to read a directory gets a permission error
        # this is just an easy way for tests to simulate bad permissions
        os.makedirs(fpath)

    with pytest.raises(PermissionError):
        local_gcp.endpoint_id



import pytest

import globus_sdk


@pytest.fixture
def groups_client():
    client = globus_sdk.GroupsClient()
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def groups_manager(groups_client):
    return globus_sdk.GroupsManager(groups_client)



import json
import os

import pytest

from globus_sdk import __version__
from globus_sdk.token_storage.legacy import SimpleJSONFileAdapter

IS_WINDOWS = os.name == "nt"


@pytest.fixture
def json_file(tmp_path):
    return tmp_path / "mydata.json"


def test_file_does_not_exist(json_file):
    adapter = SimpleJSONFileAdapter(json_file)
    assert not adapter.file_exists()


def test_file_exists(json_file):
    json_file.touch()
    adapter = SimpleJSONFileAdapter(json_file)
    assert adapter.file_exists()


def test_store(json_file, mock_response):
    adapter = SimpleJSONFileAdapter(json_file)
    assert not adapter.file_exists()
    adapter.store(mock_response)

    data = json.loads(json_file.read_text())
    assert data["globus-sdk.version"] == __version__
    assert data["by_rs"]["resource_server_1"]["access_token"] == "access_token_1"
    assert data["by_rs"]["resource_server_2"]["access_token"] == "access_token_2"


def test_get_token_data(json_file, mock_response):
    adapter = SimpleJSONFileAdapter(json_file)
    assert not adapter.file_exists()
    adapter.store(mock_response)

    data = adapter.get_token_data("resource_server_1")
    assert data["access_token"] == "access_token_1"


def test_store_and_refresh(json_file, mock_response, mock_refresh_response):
    adapter = SimpleJSONFileAdapter(json_file)
    assert not adapter.file_exists()
    adapter.store(mock_response)

    # rs1 and rs2 data was stored correctly
    data = adapter.get_token_data("resource_server_1")
    assert data["access_token"] == "access_token_1"
    data = adapter.get_token_data("resource_server_2")
    assert data["access_token"] == "access_token_2"

    # "refresh" happens, this should change rs2 but not rs1
    adapter.store(mock_refresh_response)
    data = adapter.get_token_data("resource_server_1")
    assert data["access_token"] == "access_token_1"
    data = adapter.get_token_data("resource_server_2")
    assert data["access_token"] == "access_token_2_refreshed"


@pytest.mark.xfail(IS_WINDOWS, reason="cannot set umask perms on Windows")
def test_store_perms(json_file, mock_response):
    adapter = SimpleJSONFileAdapter(json_file)
    assert not adapter.file_exists()
    adapter.store(mock_response)

    # mode|0600 should be 0600 -- meaning that those are the maximal
    # permissions given
    st_mode = json_file.stat().st_mode & 0o777  # & 777 to remove extra bits
    assert st_mode | 0o600 == 0o600



import typing as t
from pydoc import locate

from ..utils import read_sphinx_params
from .add_content_directive import AddContentDirective


class CopyParams(AddContentDirective):
    has_content = True
    required_arguments = 1
    optional_arguments = 0

    def gen_rst(self) -> t.Iterator[str]:
        source_object_name: str = self.arguments[0]
        if not source_object_name.startswith("globus_sdk"):
            source_object_name = f"globus_sdk.{source_object_name}"

        source_object = locate(source_object_name)

        content: t.Iterable[str]
        if self.content:
            content = iter(self.content)
        else:
            content = ()

        for line in content:
            if line.strip() == "<YIELD>":
                break
            yield line

        yield from read_sphinx_params(source_object.__doc__)

        for line in content:
            yield line



import pytest

import globus_sdk
from globus_sdk.scopes import Scope


@pytest.fixture
def user_app():
    client_id = "mock_client_id"
    return globus_sdk.UserApp("test-app", client_id=client_id)


@pytest.mark.parametrize(
    "scope_collection",
    ([Scope("scope1")], Scope("scope1"), "scope1", ["scope1"]),
)
def test_iter_scopes_simple(user_app, scope_collection):
    actual_list = list(user_app._iter_scopes(scope_collection))

    assert len(actual_list) == 1
    assert isinstance(actual_list[0], Scope)
    assert str(actual_list[0]) == "scope1"


@pytest.mark.parametrize(
    "scope_collection, expect_str",
    (
        (("scope1", "scope2"), "scope1 scope2"),
        (("scope1", Scope("scope2")), "scope1 scope2"),
        ((Scope("scope1"), Scope("scope2")), "scope1 scope2"),
        ((Scope("scope1"), Scope("scope2"), "scope3"), "scope1 scope2 scope3"),
        (
            (Scope("scope1"), Scope("scope2"), "scope3 scope4"),
            "scope1 scope2 scope3 scope4",
        ),
        ([Scope("scope1"), "scope2", "scope3 scope4"], "scope1 scope2 scope3 scope4"),
    ),
)
def test_iter_scopes_handles_mixed_data(user_app, scope_collection, expect_str):
    actual_list = list(user_app._iter_scopes(scope_collection))

    assert all(isinstance(scope, Scope) for scope in actual_list)
    assert _as_sorted_string(actual_list) == expect_str


def test_iter_scopes_handles_dependent_scopes(user_app):
    scope_collection = "scope1 scope2[scope3 scope4]"
    actual_list = list(user_app._iter_scopes(scope_collection))

    actual_sorted_str = _as_sorted_string(actual_list)
    # Dependent scope ordering is not guaranteed
    assert (
        actual_sorted_str == "scope1 scope2[scope3 scope4]"
        or actual_sorted_str == "scope1 scope2[scope4 scope3]"
    )


def _as_sorted_string(scope_list) -> str:
    return " ".join(sorted(str(scope) for scope in scope_list))



from .base_login_client import AuthLoginClient
from .confidential_client import ConfidentialAppAuthClient
from .native_client import NativeAppAuthClient
from .service_client import AuthClient

__all__ = (
    "AuthClient",
    "AuthLoginClient",
    "NativeAppAuthClient",
    "ConfidentialAppAuthClient",
)



from __future__ import annotations

import uuid
from datetime import datetime

from globus_sdk import GlobusError, Scope


class TokenValidationError(GlobusError):
    """The class of errors raised when a token fails validation."""


class IdentityValidationError(TokenValidationError):
    """
    The class of errors raised when a token response's identity is indeterminate or
    incorrect.
    """


class MissingIdentityError(IdentityValidationError, LookupError):
    """No identity info contained in a token response."""


class IdentityMismatchError(IdentityValidationError, ValueError):
    """The identity in a token response did not match the expected identity."""

    def __init__(
        self, message: str, stored_id: uuid.UUID | str, new_id: uuid.UUID | str
    ) -> None:
        super().__init__(message)
        self.stored_id = stored_id
        self.new_id = new_id


class MissingTokenError(TokenValidationError, LookupError):
    """No token stored for a given resource server."""

    def __init__(self, message: str, resource_server: str) -> None:
        super().__init__(message)
        self.resource_server = resource_server


class ExpiredTokenError(TokenValidationError, ValueError):
    """The token stored for a given resource server has expired."""

    def __init__(self, expires_at_seconds: int) -> None:
        expiration = datetime.fromtimestamp(expires_at_seconds)
        super().__init__(f"Token expired at {expiration.isoformat()}")
        self.expiration = expiration


class UnmetScopeRequirementsError(TokenValidationError, ValueError):
    """The token stored for a given resource server is missing required scopes."""

    def __init__(
        self, message: str, scope_requirements: dict[str, list[Scope]]
    ) -> None:
        super().__init__(message)
        # The full set of scope requirements which were evaluated.
        #   Notably this is not exclusively the unmet scope requirements.
        self.scope_requirements = scope_requirements



import json
from unittest import mock

import pytest
import requests

from globus_sdk.paging import HasNextPaginator, JSONAPIPaginator
from globus_sdk.response import GlobusHTTPResponse, IterableJSONAPIResponse
from globus_sdk.services.transfer.response import IterableTransferResponse

N = 25


class PagingSimulator:
    def __init__(self, n) -> None:
        self.n = n  # the number of simulated items

    def simulate_get(self, *args, **params):
        """
        Simulates a paginated response from a Globus API get supporting limit,
        offset, and has next page
        """
        offset = params.get("offset", 0)
        limit = params["limit"]
        data = {}  # dict that will be treated as the json data of a response
        data["offset"] = offset
        data["limit"] = limit
        # fill data field
        data["DATA"] = []
        for i in range(offset, min(self.n, offset + limit)):
            data["DATA"].append({"value": i})
        # fill has_next_page field
        data["has_next_page"] = (offset + limit) < self.n

        # make the simulated response
        response = requests.Response()
        response._content = json.dumps(data).encode()
        response.headers["Content-Type"] = "application/json"
        return IterableTransferResponse(GlobusHTTPResponse(response, mock.Mock()))


class JSONAPIPagingSimulator:

    def __init__(self, n) -> None:
        self.n = n  # the number of simulated items
        self.page_size = 10  # arbitrary page size

    def simulate_get(self, *args, **params):
        """
        Simulates a paginated response from a Globus API get supporting
        optional query_parameters and including a links object with a next
        link in the response using a simple id based page marker
        """
        query_params = params.get("query_params", {})
        marker = query_params.get("page[marker]")

        if not marker:
            marker = 0
        elif isinstance(marker, list):
            marker = int(marker[0])

        if marker > self.n:
            raise Exception("BadMarker")

        else:
            response_top_level = {
                "data": [],
            }
            for i in range(marker, min(self.n, marker + self.page_size)):
                response_top_level["data"].append(
                    {
                        "type": "foo",
                        "id": i,
                    }
                )

            # add links object with next link if there is remaining data
            if marker + self.page_size < self.n:
                next_marker = marker + self.page_size
                response_top_level["links"] = {
                    "next": f"https://foo.globus.org/foo?page[marker]={next_marker}"
                }

        # make the simulated response
        response = requests.Response()
        response._content = json.dumps(response_top_level).encode()
        response.headers["Content-Type"] = "application/json"
        return IterableJSONAPIResponse(GlobusHTTPResponse(response, mock.Mock()))


@pytest.fixture
def paging_simulator():
    return PagingSimulator(N)


@pytest.fixture
def jsonapi_paging_simulator():
    return JSONAPIPagingSimulator(N)


def test_has_next_paginator(paging_simulator):
    """
    Walk the paging simulator with HasNextPaginator and confirm the results are good
    """
    paginator = HasNextPaginator(
        paging_simulator.simulate_get,
        get_page_size=lambda x: len(x["DATA"]),
        max_total_results=1000,
        page_size=10,
        client_args=[],
        client_kwargs={},
    )

    def all_items():
        for page in paginator:
            yield from page

    # confirm results
    for item, expected in zip(all_items(), range(N)):
        assert item["value"] == expected


def test_jsonapi_paginator(jsonapi_paging_simulator):
    """
    Walk the JSONAPIPagingSimulator with JSONAPIPaginator and confirm results are good
    """
    paginator = JSONAPIPaginator(
        jsonapi_paging_simulator.simulate_get,
        client_args=[],
        client_kwargs={},
    )

    def all_items():
        for page in paginator:
            yield from page

    # confirm results
    for item, expected in zip(all_items(), range(N)):
        assert item["id"] == expected



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_DOC, TWO_HOP_TRANSFER_FLOW_ID

RESPONSES = ResponseSet(
    metadata={
        "flow_id": TWO_HOP_TRANSFER_FLOW_ID,
        "title": TWO_HOP_TRANSFER_FLOW_DOC["title"],
    },
    default=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}",
        json=TWO_HOP_TRANSFER_FLOW_DOC,
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID, FUNCTION_ID, TASK_GROUP_ID, TASK_ID, TASK_ID_2

REQUEST_ID = "5158de19-10b5-4deb-9d87-a86c1dec3460"

SUBMIT_RESPONSE = {
    "request_id": REQUEST_ID,
    "task_group_id": TASK_GROUP_ID,
    "endpoint_id": ENDPOINT_ID,
    "tasks": {
        FUNCTION_ID: [TASK_ID, TASK_ID_2],
    },
}

RESPONSES = ResponseSet(
    metadata={
        "endpoint_id": ENDPOINT_ID,
        "function_id": FUNCTION_ID,
        "task_id": TASK_ID,
        "task_id_2": TASK_ID_2,
        "task_group_id": TASK_GROUP_ID,
        "request_id": REQUEST_ID,
    },
    default=RegisteredResponse(
        service="compute",
        path=f"/v3/endpoints/{ENDPOINT_ID}/submit",
        method="POST",
        json=SUBMIT_RESPONSE,
    ),
)



from __future__ import annotations

import uuid
from collections import defaultdict, namedtuple
from dataclasses import dataclass, field
from datetime import datetime, timedelta

from globus_sdk.scopes import Scope, ScopeParser
from globus_sdk.scopes.consents import Consent, ConsentForest

ScopeRepr = namedtuple("Scope", ["id", "name"])


@dataclass
class ConsentTest(Consent):
    """
    A convenience Consent data subclass with default values for most fields to make
      test case definition less verbose.

    Required fields: client, scope, scope_name
    """

    client: uuid.UUID | str
    scope: uuid.UUID | str
    scope_name: str
    id: int = field(default_factory=lambda: uuid.uuid1().int)
    effective_identity: uuid.UUID | str = str(uuid.uuid4())
    dependency_path: list[int] = field(default_factory=list)
    created: datetime = field(
        default_factory=lambda: datetime.now() - timedelta(days=1)
    )
    updated: datetime = field(
        default_factory=lambda: datetime.now() - timedelta(days=1)
    )
    last_used: datetime = field(default_factory=datetime.now)
    status: str = "approved"
    allows_refresh: bool = True
    auto_approved: bool = False
    atomically_revocable: bool = False

    def __post_init__(self):
        # Append self to the dependency path if it's not already there
        if not self.dependency_path or self.dependency_path[-1] != self.id:
            self.dependency_path.append(self.id)

    @classmethod
    def of(
        cls,
        client: str,
        scope: ScopeRepr,
        *,
        parent: ConsentTest | None = None,
        **kwargs,
    ) -> ConsentTest:
        return cls(
            client=client,
            scope=scope.id,
            scope_name=scope.name,
            dependency_path=list(parent.dependency_path) if parent else [],
            **kwargs,
        )


def make_consent_forest(scopes: list[str | Scope] | str | Scope) -> ConsentForest:
    """
    Creates a consent forest from a list of scope strings or scope objects.
    Client and Scope IDs are generated at random.
    """
    _scopes = _normalize_scopes(scopes)
    scope_id_mapping: dict[str, str] = defaultdict(lambda: str(uuid.uuid4()))
    consents = []
    for scope in _scopes:
        consents.extend(_generate_consents(scope, scope_id_mapping))
    return ConsentForest(consents)


def _normalize_scopes(scopes: list[str | Scope] | str | Scope) -> list[Scope]:
    if isinstance(scopes, Scope):
        return [scopes]
    elif isinstance(scopes, str):
        return ScopeParser.parse(scopes)
    else:
        to_return = []
        for scope in scopes:
            to_return.extend(_normalize_scopes(scope))
        return to_return


def _generate_consents(
    scope: Scope, scope_id_mapping: dict[str, str], parent: ConsentTest | None = None
) -> list[Consent]:
    """Generates a list of consents for a scope and its children."""
    consents = []
    client_id = str(uuid.uuid4())
    scope_string = scope.scope_string
    scope_id = scope_id_mapping[scope_string]
    consent = ConsentTest.of(
        client_id, ScopeRepr(scope_id, scope_string), parent=parent
    )
    consents.append(consent)
    for dependent_scope in scope.dependencies:
        consents.extend(
            _generate_consents(dependent_scope, scope_id_mapping, parent=consent)
        )
    return consents



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

CREDENTIAL_IDS = [
    "af43d884-64a1-4414-897a-680c32374439",
    "c96b8f70-1448-46db-89af-292623c93ee4",
]

RESPONSES = ResponseSet(
    metadata={"ids": CREDENTIAL_IDS},
    default=RegisteredResponse(
        service="gcs",
        path="/user_credentials",
        method="GET",
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "success",
            "data": [
                {
                    "DATA_TYPE": "user_credential#1.0.0",
                    "connector_id": "145812c8-decc-41f1-83cf-bb2a85a2a70b",
                    "display_name": "posix_credential",
                    "id": CREDENTIAL_IDS[0],
                    "identity_id": "948847d4-ffcc-4ae0-ba3a-a4c88d480159",
                    "invalid": False,
                    "policies": {"DATA_TYPE": "posix_user_credential_policies#1.0.0"},
                    "provisioned": False,
                    "storage_gateway_id": "82247cc9-3208-4d71-bd7f-1b8798c95e6b",
                    "username": "testuser",
                },
                {
                    "DATA_TYPE": "user_credential#1.0.0",
                    "connector_id": "7643e831-5f6c-4b47-a07f-8ee90f401d23",
                    "display_name": "s3_credential",
                    "id": CREDENTIAL_IDS[1],
                    "identity_id": "948847d4-ffcc-4ae0-ba3a-a4c88d480159",
                    "invalid": False,
                    "policies": {
                        "DATA_TYPE": "s3_user_credential_policies#1.0.0",
                        "s3_key_id": "key_id",
                        "s3_secret_key": "key_secret",
                    },
                    "provisioned": True,
                    "storage_gateway_id": "99aab7ac-8fde-40e2-b6db-44de8e59597a",
                    "username": "testuser",
                },
            ],
            "detail": "success",
            "has_next_page": False,
            "http_response_code": 200,
        },
        metadata={"ids": CREDENTIAL_IDS},
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import GROUP_ID, SUBSCRIPTION_ID

RESPONSES = ResponseSet(
    metadata={"group_id": GROUP_ID, "subscription_id": SUBSCRIPTION_ID},
    default=RegisteredResponse(
        service="groups",
        path=f"/v2/groups/{GROUP_ID}/subscription_admin_verified",
        method="PUT",
        json={
            "group_id": GROUP_ID,
            "subscription_admin_verified_id": SUBSCRIPTION_ID,
        },
    ),
)



"""
Common use helpers and utilities for all tests to leverage.
Not so disorganized as a "utils" module and not so refined as a public package.
"""

from .consents import ConsentTest, ScopeRepr, make_consent_forest
from .constants import GO_EP1_ID, GO_EP2_ID
from .globus_responses import register_api_route, register_api_route_fixture_file
from .response_mock import PickleableMockResponse

__all__ = [
    "ConsentTest",
    "GO_EP1_ID",
    "GO_EP2_ID",
    "make_consent_forest",
    "PickleableMockResponse",
    "register_api_route",
    "register_api_route_fixture_file",
    "ScopeRepr",
]



import pathlib
from unittest import mock

import pytest

from globus_sdk.transport import RequestsTransport, RetryConfig, RetryContext
from globus_sdk.transport.retry_config import _exponential_backoff


def _linear_backoff(ctx: RetryContext) -> float:
    if ctx.backoff is not None:
        return ctx.backoff
    return 0.5 * (2**ctx.attempt)


ca_bundle_file = pathlib.Path(__file__).parent.parent.absolute() / "CA-Bundle.cert"
ca_bundle_directory = ca_bundle_file.parent
ca_bundle_non_existent = ca_bundle_directory / "bogus.bogus"


@pytest.mark.parametrize(
    "param_name, init_value, tune_value",
    [
        ("verify_ssl", True, True),
        ("verify_ssl", True, False),
        ("verify_ssl", True, ca_bundle_file),
        ("verify_ssl", True, str(ca_bundle_file)),
        ("verify_ssl", True, ca_bundle_directory),
        ("verify_ssl", True, str(ca_bundle_directory)),
        ("verify_ssl", True, ca_bundle_non_existent),
        ("verify_ssl", True, str(ca_bundle_non_existent)),
        ("http_timeout", 60, 120),
    ],
)
def test_transport_tuning(param_name, init_value, tune_value):
    expected_value = (
        str(tune_value) if isinstance(tune_value, pathlib.Path) else tune_value
    )
    init_kwargs = {param_name: init_value}
    transport = RequestsTransport(**init_kwargs)

    assert getattr(transport, param_name) == init_value

    tune_kwargs = {param_name: tune_value}
    with transport.tune(**tune_kwargs):
        assert getattr(transport, param_name) == expected_value

    assert getattr(transport, param_name) == init_value


@pytest.mark.parametrize(
    "param_name, init_value, tune_value",
    [
        ("backoff", _exponential_backoff, _linear_backoff),
        ("max_sleep", 10, 10),
        ("max_sleep", 10, 1),
        ("max_retries", 0, 5),
        ("max_retries", 10, 0),
    ],
)
def test_retry_tuning(param_name, init_value, tune_value):
    init_kwargs = {param_name: init_value}
    config = RetryConfig(**init_kwargs)

    assert getattr(config, param_name) == init_value

    tune_kwargs = {param_name: tune_value}
    with config.tune(**tune_kwargs):
        assert getattr(config, param_name) == tune_value

    assert getattr(config, param_name) == init_value


def test_transport_can_manipulate_user_agent():
    transport = RequestsTransport()

    # starting state -- the user agent is visible and equivalent in two places
    assert "User-Agent" in transport.headers
    assert transport.headers["User-Agent"] == transport.user_agent

    # setting it slash-appends it to the base agent (in both places)
    transport.user_agent = "frobulator-1.0"
    assert (
        transport.headers["User-Agent"] == f"{transport.BASE_USER_AGENT}/frobulator-1.0"
    )
    assert transport.user_agent == f"{transport.BASE_USER_AGENT}/frobulator-1.0"

    # deleting it doesn't clear the property
    # but accessing the property doesn't add it back into the dict
    del transport.headers["User-Agent"]
    assert transport.user_agent == f"{transport.BASE_USER_AGENT}/frobulator-1.0"
    assert "User-Agent" not in transport.headers

    # but updating it will add it back
    transport.user_agent = "demuddler-2.2"
    assert (
        transport.headers["User-Agent"] == f"{transport.BASE_USER_AGENT}/demuddler-2.2"
    )
    assert transport.user_agent == f"{transport.BASE_USER_AGENT}/demuddler-2.2"


def test_transport_can_manipulate_client_info():
    transport = RequestsTransport()

    # starting state -- the clientinfo is visible in headers
    # and a fake product is not
    assert "X-Globus-Client-Info" in transport.headers
    assert "frobulator" not in transport.headers["X-Globus-Client-Info"]

    # we can add said product to the header
    transport.globus_client_info.add({"product": "frobulator", "version": "1.0"})
    assert "product=frobulator,version=1.0" in transport.headers["X-Globus-Client-Info"]

    # clearing the client info removes the header
    transport.globus_client_info.clear()
    assert "X-Globus-Client-Info" not in transport.headers

    # adding a product re-adds the header
    transport.globus_client_info.add({"product": "demuddler", "version": "8.8"})
    assert "product=demuddler,version=8.8" in transport.headers["X-Globus-Client-Info"]


def test_double_clear_of_client_info_is_allowed():
    # clear the client info twice -- catching any potential bugs with `del ...` or
    # similar assumptions that the value is present
    transport = RequestsTransport()
    transport.globus_client_info.clear()
    transport.globus_client_info.clear()
    assert "X-Globus-Client-Info" not in transport.headers


def test_transport_close_closes_session():
    # there's no easy way to check if a real requests.Session object is closed,
    # so just check that `close()` was called
    transport = RequestsTransport()
    with mock.patch.object(transport, "session") as mocked_session:
        transport.close()
        mocked_session.close.assert_called_once_with()



from globus_sdk.response import IterableResponse


class GetClientCredentialsResponse(IterableResponse):
    """
    Response class specific to the Get Credentials API

    Provides iteration on the "credentials" array in the response.
    """

    default_iter_key = "credentials"



import unittest.mock
import uuid

import pytest

import globus_sdk

CLIENT_ID_UUID = uuid.uuid4()
CLIENT_ID_STR = str(CLIENT_ID_UUID)


@pytest.mark.parametrize(
    "client_type",
    (
        globus_sdk.AuthLoginClient,
        globus_sdk.ConfidentialAppAuthClient,
        globus_sdk.NativeAppAuthClient,
    ),
)
@pytest.mark.parametrize(
    "pass_value", (CLIENT_ID_STR, CLIENT_ID_UUID), ids=("str", "uuid")
)
def test_can_use_uuid_or_str_for_client_id(client_type, pass_value):
    if client_type in (globus_sdk.AuthLoginClient, globus_sdk.NativeAppAuthClient):
        client = client_type(client_id=pass_value)
    elif client_type is globus_sdk.ConfidentialAppAuthClient:
        client = globus_sdk.ConfidentialAppAuthClient(pass_value, "bogus_secret")
    else:
        raise NotImplementedError

    assert client.client_id == CLIENT_ID_STR


def test_native_app_auth_client_rejects_authorizer():
    authorizer = unittest.mock.Mock()
    with pytest.raises(TypeError):
        globus_sdk.NativeAppAuthClient(CLIENT_ID_UUID, authorizer=authorizer)


def test_confidential_app_auth_client_rejects_authorizer():
    authorizer = unittest.mock.Mock()
    with pytest.raises(TypeError):
        globus_sdk.ConfidentialAppAuthClient(
            CLIENT_ID_UUID, "foo-secret", authorizer=authorizer
        )
